"""Shared utilities for copilot functionality."""

from typing import List, Optional

from rasa.builder.copilot.models import (
    ChatMessage,
    CopilotChatMessage,
    ResponseCategory,
    UserChatMessage,
)


def filter_chat_history_messages(
    chat_history: List[ChatMessage],
    excluded_response_categories: Optional[List[ResponseCategory]] = None,
) -> List[UserChatMessage | CopilotChatMessage]:
    """Filter and convert chat history messages.

    This method will only return UserChatMessage and CopilotChatMessage messages.

    Args:
        chat_history: List of chat messages to filter and convert.
        excluded_response_categories: List of response categories to exclude.
            Defaults to [ResponseCategory.GUARDRAILS_POLICY_VIOLATION].
        allowed_roles: List of roles to keep. Defaults to ["user", "copilot"].

    Returns:
        List of messages in OpenAI format.
    """
    filtered_messages: List[UserChatMessage | CopilotChatMessage] = []

    for message in chat_history:
        # Check if message should be excluded based on response category
        if (
            excluded_response_categories is not None
            and message.response_category in excluded_response_categories
        ):
            continue

        # Only process UserChatMessage and CopilotChatMessage
        if not isinstance(message, (UserChatMessage, CopilotChatMessage)):
            continue

        filtered_messages.append(message)

    return filtered_messages



from typing import Any, Dict, Optional

import tensorflow as tf
from tensorflow.types.experimental import TensorLike

# original code taken from
# https://github.com/tensorflow/addons/blob/f30df4322b5580b3e5946530a60f7126035dd73b/tensorflow_addons/metrics/f_scores.py
# (modified to our neeeds)


class FBetaScore(tf.keras.metrics.Metric):
    r"""Computes F-Beta score.

    It is the weighted harmonic mean of precision
    and recall. Output range is `[0, 1]`. Works for
    both multi-class and multi-label classification.

    $$
    F_{\beta} = (1 + \beta^2) * \frac{\textrm{precision} * \textrm{recall}}
                                  {(\beta^2 \cdot \textrm{precision}) + \textrm{recall}}
    $$

    Args:
        num_classes: Number of unique classes in the dataset.
        average: Type of averaging to be performed on data.
            Acceptable values are `None`, `micro`, `macro` and
            `weighted`. Default value is None.
        beta: Determines the weight of precision and recall
            in harmonic mean. Determines the weight given to the
            precision and recall. Default value is 1.
        threshold: Elements of `y_pred` greater than threshold are
            converted to be 1, and the rest 0. If threshold is
            None, the argmax is converted to 1, and the rest 0.
        name: (Optional) String name of the metric instance.
        dtype: (Optional) Data type of the metric result.

    Returns:
        F-Beta Score: float.

    Raises:
        ValueError: If the `average` has values other than
        `[None, 'micro', 'macro', 'weighted']`.

        ValueError: If the `beta` value is less than or equal
        to 0.

    `average` parameter behavior:

        None: Scores for each class are returned.

        micro: True positivies, false positives and
            false negatives are computed globally.

        macro: True positivies, false positives and
            false negatives are computed for each class
            and their unweighted mean is returned.

        weighted: Metrics are computed for each class
            and returns the mean weighted by the
            number of true instances in each class.

    Usage:

    >>> metric = tfa.metrics.FBetaScore(num_classes=3, beta=2.0, threshold=0.5)
    >>> y_true = np.array([[1, 1, 1],
    ...                    [1, 0, 0],
    ...                    [1, 1, 0]], np.int32)
    >>> y_pred = np.array([[0.2, 0.6, 0.7],
    ...                    [0.2, 0.6, 0.6],
    ...                    [0.6, 0.8, 0.0]], np.float32)
    >>> metric.update_state(y_true, y_pred)
    >>> result = metric.result()
    >>> result.numpy()
    array([0.3846154 , 0.90909094, 0.8333334 ], dtype=float32)
    """

    def __init__(
        self,
        num_classes: TensorLike,
        average: Optional[str] = None,
        beta: TensorLike = 1.0,
        threshold: Optional[TensorLike] = None,
        name: str = "fbeta_score",
        dtype: Any = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(name=name, dtype=dtype)

        if average not in (None, "micro", "macro", "weighted"):
            raise ValueError(
                "Unknown average type. Acceptable values "
                "are: [None, 'micro', 'macro', 'weighted']"
            )

        if not isinstance(beta, float):
            raise TypeError("The value of beta should be a python float")

        if beta <= 0.0:
            raise ValueError("beta value should be greater than zero")

        if threshold is not None:
            if not isinstance(threshold, float):
                raise TypeError("The value of threshold should be a python float")
            if threshold > 1.0 or threshold <= 0.0:
                raise ValueError("threshold should be between 0 and 1")

        self.num_classes = num_classes
        self.average = average
        self.beta = beta
        self.threshold = threshold
        self.axis = None
        self.init_shape = []

        if self.average != "micro":
            self.axis = 0
            self.init_shape = [self.num_classes]

        def _zero_wt_init(name: Any) -> Any:
            return self.add_weight(
                name=name, shape=self.init_shape, initializer="zeros", dtype=self.dtype
            )

        self.true_positives = _zero_wt_init("true_positives")
        self.false_positives = _zero_wt_init("false_positives")
        self.false_negatives = _zero_wt_init("false_negatives")
        self.weights_intermediate = _zero_wt_init("weights_intermediate")

    def update_state(
        self,
        y_true: TensorLike,
        y_pred: TensorLike,
        sample_weight: Optional[TensorLike] = None,
    ) -> None:
        if self.threshold is None:
            threshold = tf.reduce_max(y_pred, axis=-1, keepdims=True)
            # make sure [0, 0, 0] doesn't become [1, 1, 1]
            # Use abs(x) > eps, instead of x != 0 to check for zero
            y_pred = tf.logical_and(y_pred >= threshold, tf.abs(y_pred) > 1e-12)
        else:
            y_pred = y_pred > self.threshold

        y_true = tf.cast(y_true, self.dtype)
        y_pred = tf.cast(y_pred, self.dtype)

        def _weighted_sum(
            val: TensorLike, sample_weight: Optional[TensorLike]
        ) -> TensorLike:
            if sample_weight is not None:
                val = tf.math.multiply(val, tf.expand_dims(sample_weight, 1))
            return tf.reduce_sum(val, axis=self.axis)

        self.true_positives.assign_add(_weighted_sum(y_pred * y_true, sample_weight))
        self.false_positives.assign_add(
            _weighted_sum(y_pred * (1 - y_true), sample_weight)
        )
        self.false_negatives.assign_add(
            _weighted_sum((1 - y_pred) * y_true, sample_weight)
        )
        self.weights_intermediate.assign_add(_weighted_sum(y_true, sample_weight))

    def result(self) -> TensorLike:
        precision = tf.math.divide_no_nan(
            self.true_positives, self.true_positives + self.false_positives
        )
        recall = tf.math.divide_no_nan(
            self.true_positives, self.true_positives + self.false_negatives
        )

        mul_value = precision * recall
        add_value = (tf.math.square(self.beta) * precision) + recall
        mean = tf.math.divide_no_nan(mul_value, add_value)
        f1_score = mean * (1 + tf.math.square(self.beta))

        if self.average == "weighted":
            weights = tf.math.divide_no_nan(
                self.weights_intermediate, tf.reduce_sum(self.weights_intermediate)
            )
            f1_score = tf.reduce_sum(f1_score * weights)

        elif self.average is not None:  # [micro, macro]
            f1_score = tf.reduce_mean(f1_score)

        return f1_score

    def get_config(self) -> Dict[str, Any]:
        """Returns the serializable config of the metric."""
        config = {
            "num_classes": self.num_classes,
            "average": self.average,
            "beta": self.beta,
            "threshold": self.threshold,
        }

        base_config = super().get_config()
        return {**base_config, **config}

    def reset_state(self) -> None:
        reset_value = tf.zeros(self.init_shape, dtype=self.dtype)
        # In Keras 3.x, self.variables contains string names, not variable objects,
        # so each metric variable is reset using assign() instead of K.batch_set_value()
        self.true_positives.assign(reset_value)
        self.false_positives.assign(reset_value)
        self.false_negatives.assign(reset_value)
        self.weights_intermediate.assign(reset_value)

    def reset_states(self) -> None:
        # Backwards compatibility alias of `reset_state`. New classes should
        # only implement `reset_state`.
        # Required in Tensorflow < 2.5.0
        return self.reset_state()


class F1Score(FBetaScore):
    r"""Computes F-1 Score.

    It is the harmonic mean of precision and recall.
    Output range is `[0, 1]`. Works for both multi-class
    and multi-label classification.

    $$
    F_1 = 2 \cdot \frac{\textrm{precision} \cdot \textrm{recall}}{\textrm{precision}
          + \textrm{recall}}
    $$

    Args:
        num_classes: Number of unique classes in the dataset.
        average: Type of averaging to be performed on data.
            Acceptable values are `None`, `micro`, `macro`
            and `weighted`. Default value is None.
        threshold: Elements of `y_pred` above threshold are
            considered to be 1, and the rest 0. If threshold is
            None, the argmax is converted to 1, and the rest 0.
        name: (Optional) String name of the metric instance.
        dtype: (Optional) Data type of the metric result.

    Returns:
        F-1 Score: float.

    Raises:
        ValueError: If the `average` has values other than
        [None, 'micro', 'macro', 'weighted'].

    `average` parameter behavior:
        None: Scores for each class are returned

        micro: True positivies, false positives and
            false negatives are computed globally.

        macro: True positivies, false positives and
            false negatives are computed for each class
            and their unweighted mean is returned.

        weighted: Metrics are computed for each class
            and returns the mean weighted by the
            number of true instances in each class.

    Usage:

    >>> metric = tfa.metrics.F1Score(num_classes=3, threshold=0.5)
    >>> y_true = np.array([[1, 1, 1],
    ...                    [1, 0, 0],
    ...                    [1, 1, 0]], np.int32)
    >>> y_pred = np.array([[0.2, 0.6, 0.7],
    ...                    [0.2, 0.6, 0.6],
    ...                    [0.6, 0.8, 0.0]], np.float32)
    >>> metric.update_state(y_true, y_pred)
    >>> result = metric.result()
    >>> result.numpy()
    array([0.5      , 0.8      , 0.6666667], dtype=float32)
    """

    def __init__(
        self,
        num_classes: TensorLike,
        average: Optional[str] = None,
        threshold: Optional[TensorLike] = None,
        name: str = "f1_score",
        dtype: Any = None,
    ):
        super().__init__(num_classes, average, 1.0, threshold, name=name, dtype=dtype)

    def get_config(self) -> Dict[str, Any]:
        base_config = super().get_config()
        del base_config["beta"]
        return base_config



"""Fundamental types for authentication protocols."""

from enum import Enum


class AgentAuthType(Enum):
    """An Enum class that represents the supported authentication protocol types."""

    API_KEY = "api_key"
    OAUTH2 = "oauth2"
    BEARER_TOKEN = "bearer_token"
    CUSTOM = "custom"



"""Service for generating Rasa projects from prompts."""

import asyncio
import json
import os
import re
import shutil
import subprocess
from contextlib import asynccontextmanager
from copy import deepcopy
from pathlib import Path
from typing import Any, AsyncGenerator, Dict, Generator, List, Optional, Tuple, cast

import importlib_resources
import openai
import structlog
from jinja2 import Template
from openai.types.chat import ChatCompletion

from rasa.builder import config
from rasa.builder.config import PROJECT_GENERATION_TIMEOUT
from rasa.builder.copilot import get_copilot_class
from rasa.builder.copilot.constants import (
    DEFAULT_COMMIT_MESSAGE,
)
from rasa.builder.copilot.models import (
    CopilotContext,
    FileContent,
    InternalCopilotRequestChatMessage,
    LogContent,
    ResponseCategory,
    UsageStatistics,
)
from rasa.builder.exceptions import (
    LLMGenerationError,
    ProjectGenerationError,
    ValidationError,
)
from rasa.builder.git_service import DEFAULT_COMMIT_INFO, GitService
from rasa.builder.logging_utils import capture_exception_with_context
from rasa.builder.models import BotFiles, GitCommitInfo
from rasa.builder.project_generator.project_utils import (
    bot_file_paths,
    get_bot_files,
    is_restricted_path,
    path_relative_to_project,
    unsafe_write_to_bot_files,
    validate_file_content_transitions,
)
from rasa.builder.project_info import ProjectInfo, ensure_first_used, load_project_info
from rasa.builder.telemetry.langfuse.commit_langfuse_telemetry import (
    CommitMessageGenerationLangfuseTelemetry,
)
from rasa.builder.telemetry.langfuse.langfuse_compat import observe
from rasa.builder.telemetry.langfuse.prompt_to_bot_langfuse_telemetry import (
    PromptToBotLangfuseTelemetry,
)
from rasa.builder.telemetry.langfuse.welcome_langfuse_telemetry import (
    WelcomeMessageGenerationLangfuseTelemetry,
)
from rasa.builder.template_cache import copy_cache_for_template_if_available
from rasa.builder.training_service import TrainingInput
from rasa.builder.validation_service import validate_project
from rasa.cli.scaffold import ProjectTemplateName, create_initial_project
from rasa.shared.constants import (
    DOMAIN_SCHEMA_FILE,
    PACKAGE_NAME,
    RESPONSES_SCHEMA_FILE,
    ROLE_ASSISTANT,
    ROLE_SYSTEM,
    ROLE_USER,
)
from rasa.shared.core.flows import yaml_flows_io
from rasa.shared.core.flows.yaml_flows_io import FLOWS_SCHEMA_FILE
from rasa.shared.importers.importer import TrainingDataImporter
from rasa.shared.utils.io import read_json_file
from rasa.shared.utils.yaml import dump_obj_as_yaml_to_string, read_schema_file
from rasa.utils.io import InvalidPathException

structlogger = structlog.get_logger()

BULLET_POINT_REGEX = re.compile(r"^-\s+\*[^*]+\*\s*$")


class ProjectGenerator:
    """Service for generating Rasa projects from skill descriptions."""

    def __init__(self, project_folder: str) -> None:
        """Initialize the project generator with a folder for file persistence.

        Args:
            project_folder: Path to the folder where project files will be stored
        """
        self._client: Optional[openai.AsyncOpenAI] = None
        self.project_folder = Path(project_folder)
        self.project_folder.mkdir(parents=True, exist_ok=True)
        self.git_service = GitService(project_folder)
        self._skill_to_bot_system_prompt = Template(
            importlib_resources.read_text(  # type: ignore[no-untyped-call]
                "rasa.builder.project_generator.prompts",
                "skill_to_bot_system_prompt.jinja2",
            )
        )
        self._skill_to_bot_user_prompt = Template(
            importlib_resources.read_text(  # type: ignore[no-untyped-call]
                "rasa.builder.project_generator.prompts",
                "skill_to_bot_user_request_prompt.jinja2",
            )
        )
        self._error_feedback_template = Template(
            importlib_resources.read_text(  # type: ignore[no-untyped-call]
                "rasa.builder.project_generator.prompts",
                "skill_to_bot_error_feedback_prompt.jinja2",
            )
        )
        self._welcome_message_prompt_template = Template(
            importlib_resources.read_text(  # type: ignore[no-untyped-call]
                "rasa.builder.copilot.prompts",
                "welcome_message_prompt.jinja2",
            )
        )
        self._commit_message_prompt_template = Template(
            importlib_resources.read_text(  # type: ignore[no-untyped-call]
                "rasa.builder.copilot.prompts",
                "commit_message_prompt.jinja2",
            )
        )

        # Initialize the lists for storing the generated project data and usage
        # statistics. These are used to create error feedback messages for the next
        # attempt. Kind of like a conversation history.
        self._generated_project_data_attempts: List[Dict[str, Any]] = []
        self._usage_statistics_attempts: List[UsageStatistics] = []

        # Get the domain and flows schemas for structuring the LLM response with
        # the structured output.
        self._domain_schema = self._get_domain_schema()
        self._flows_schema = self._get_flows_schema()

        # Retrieved documentation from InKeep AI. We make it static to make the most
        # use of the input tokens caching.
        self._flow_documentation = self._get_flow_documentation()
        self._domain_documentation = self._get_domain_documentation()
        self._custom_actions_documentation = self._get_custom_actions_documentation()

        # Migrate existing projects to Git if needed
        self.migrate_git_repository_if_needed()

    @asynccontextmanager
    async def _get_client(self) -> AsyncGenerator[openai.AsyncOpenAI, None]:
        """Get or lazy create OpenAI client with proper resource management."""
        if self._client is None:
            self._client = openai.AsyncOpenAI()

        try:
            yield self._client
        except Exception as e:
            structlogger.error("project_generator.llm_client_error", error=str(e))
            raise

    def _get_domain_schema(self) -> Dict[str, Any]:
        """Return a modified domain schema dictionary for project generation."""
        domain_schema = deepcopy(
            read_schema_file(DOMAIN_SCHEMA_FILE, PACKAGE_NAME, False)
        )

        if not isinstance(domain_schema, dict):
            raise ValueError("Domain schema is not a dictionary")

        # Remove parts not needed for CALM bots
        unnecessary_keys = ["intents", "entities", "forms", "config", "session_config"]
        mapping_obj = domain_schema.get("mapping")
        if not isinstance(mapping_obj, dict):
            raise ValueError("Domain schema mapping is not a dictionary")
        mapping = cast(Dict[str, Any], mapping_obj)
        for key in unnecessary_keys:
            mapping.pop(key, None)

        # Remove problematic slot mappings
        slots_obj = mapping.get("slots")
        if isinstance(slots_obj, dict):
            slots_dict = cast(Dict[str, Any], slots_obj)
            slots_mapping_obj = cast(
                Optional[Dict[str, Any]], slots_dict.get("mapping")
            )
            if slots_mapping_obj:
                regex_slot_obj = cast(
                    Optional[Dict[str, Any]],
                    slots_mapping_obj.get("regex;([A-Za-z]+)"),
                )
                if regex_slot_obj:
                    slot_mapping_obj = cast(
                        Optional[Dict[str, Any]], regex_slot_obj.get("mapping")
                    )
                    if slot_mapping_obj:
                        slot_mapping_obj.pop("mappings", None)
                        slot_mapping_obj.pop("validation", None)

        # Add responses schema
        responses_schema_data = read_schema_file(
            RESPONSES_SCHEMA_FILE, PACKAGE_NAME, False
        )
        if not isinstance(responses_schema_data, dict):
            raise ValueError("Responses schema file is not a dictionary")
        responses_schema_raw = cast(
            Optional[Dict[str, Any]],
            responses_schema_data.get("schema;responses"),
        )
        if responses_schema_raw is None:
            raise ValueError("Responses schema is not a dictionary")
        mapping["responses"] = responses_schema_raw

        return domain_schema

    def _get_flows_schema(self) -> Dict[str, Any]:
        """Return a modified flows schema dictionary for project generation."""
        schema_file = str(
            importlib_resources.files(PACKAGE_NAME).joinpath(FLOWS_SCHEMA_FILE)
        )
        flows_schema = deepcopy(read_json_file(schema_file))
        flows_schema["$defs"]["flow"]["properties"].pop("nlu_trigger", None)
        return flows_schema

    def _get_flow_documentation(self) -> str:
        """Get the flow documentation."""
        return importlib_resources.read_text(  # type: ignore[no-untyped-call]
            "rasa.builder.project_generator.prompts",
            "flow_documentation.json",
        )

    def _get_domain_documentation(self) -> str:
        """Get the domain documentation."""
        return importlib_resources.read_text(  # type: ignore[no-untyped-call]
            "rasa.builder.project_generator.prompts",
            "domain_documentation.json",
        )

    def _get_custom_actions_documentation(self) -> str:
        """Get the custom actions documentation."""
        return importlib_resources.read_text(  # type: ignore[no-untyped-call]
            "rasa.builder.project_generator.prompts",
            "custom_actions_documentation.json",
        )

    @property
    def project_info(self) -> ProjectInfo:
        """Get the project info."""
        return load_project_info(self.project_folder)

    def is_empty(self) -> bool:
        """Check if the project folder is empty.

        Excluding hidden paths.
        """
        return not any(
            file.is_file()
            for file in self.project_folder.iterdir()
            if not file.name.startswith(".")
        )

    async def init_from_template(self, template: ProjectTemplateName) -> str:
        """Create the initial project files."""
        self.cleanup()
        create_initial_project(self.project_folder.as_posix(), template)
        # If a local cache for this template exists, copy it into the project.
        # We no longer download here to avoid blocking project creation.
        await copy_cache_for_template_if_available(template, self.project_folder)
        # needs to happen after caching, as we download/copy .rasa and that would
        # overwrite the project info file in .rasa
        ensure_first_used(self.project_folder)
        self._ensure_git_repository()

        # Create initial Git commit
        return await self.git_service.commit_changes(
            DEFAULT_COMMIT_INFO.model_copy(
                update={"message": f"Initialize project from {template.value} template"}
            )
        )

    async def generate_project_with_retries(
        self,
        skill_description: str,
        template: ProjectTemplateName,
        max_retries: Optional[int] = None,
    ) -> Tuple[int, str]:
        """Generate a Rasa project with retry logic for validation failures.

        Args:
            skill_description: Natural language description of the skill
            template: Project template to use for the initial project
            max_retries: Maximum number of retry attempts

        Returns:
            A tuple containing the
            - number of attempts taken
            - the commit SHA of the generated project

        Raises:
            ProjectGenerationError: If generation fails after all retries
        """
        if max_retries is None:
            max_retries = config.PROJECT_GENERATION_MAX_RETRIES

        # if ever we do not use the template, we need to ensure first_used is set
        # separately!
        commit_sha = await self.init_from_template(template)
        self._generated_project_data_attempts = []
        self._usage_statistics_attempts = []
        project_data = self._get_bot_data_for_llm()
        system_message = self._create_system_message()
        user_request_message = self._create_user_request_message(
            skill_description, project_data
        )
        initial_messages = [system_message, user_request_message]
        error_feedback_messages: List[Dict[str, Any]] = []

        attempts_left = max_retries
        while attempts_left > 0:
            try:
                commit_sha = await self._attempt_generation(
                    initial_messages=initial_messages,
                    error_feedback_messages=error_feedback_messages,
                    attempts_left=attempts_left,
                    max_retries=max_retries,
                )
                return max_retries - attempts_left, commit_sha

            # The generation attempt failed due to validation errors
            except ValidationError as e:
                if (attempts_left := attempts_left - 1) <= 0:
                    return max_retries, commit_sha
                    raise ProjectGenerationError(
                        f"Failed to generate valid Rasa project: {e}", max_retries
                    ) from e

                # Use the last generated project data, or fall back to original
                # if list is empty
                project_data_for_feedback = (
                    self._generated_project_data_attempts[-1]
                    if self._generated_project_data_attempts
                    else project_data
                )

                # Get copilot guidance for the error
                copilot_guidance = await self._get_copilot_error_guidance(
                    error=e, project_data=project_data_for_feedback
                )

                # Create error feedback message
                error_feedback_message = self._create_error_feedback_messages(
                    project_data=project_data_for_feedback,
                    error=e,
                    copilot_guidance=copilot_guidance,
                )
                error_feedback_messages.extend(error_feedback_message)

            # The generation attempt failed due to other errors
            except Exception as e:
                if (attempts_left := attempts_left - 1) <= 0:
                    raise ProjectGenerationError(
                        f"Failed to generate Rasa project: {e}", max_retries
                    ) from e
                # Reset error feedback messages for non-validation errors
                error_feedback_messages = []

        # This should never be reached, but satisfies type checker
        raise ProjectGenerationError(
            "Failed to generate Rasa project: exhausted all retry attempts", max_retries
        )

    @observe()
    async def _attempt_generation(
        self,
        initial_messages: List[Dict[str, Any]],
        error_feedback_messages: List[Dict[str, Any]],
        attempts_left: int,
        max_retries: int,
    ) -> str:
        """Attempt to generate and validate a project in a single try."""
        generated_project_data: Optional[Dict[str, Any]] = None
        bot_files: Optional[Dict[str, Optional[str]]] = None

        try:
            # Generate the project data
            generated_project_data = await self.generate_response(
                initial_messages=initial_messages,
                error_feedback_messages=error_feedback_messages,
            )
            # Append the generated project data to the list of attempts. This is used
            # create error feedback messages for the next attempt
            self._generated_project_data_attempts.append(generated_project_data)
            # Update the bot files from the generated project data
            commit_sha = await self._update_bot_files_from_llm_response(
                generated_project_data
            )

            bot_files = self.get_bot_files()
            structlogger.info(
                "project_generator.generated_project",
                attempts_left=attempts_left,
                max_retries=max_retries,
                files=list(bot_files.keys()),
            )

            await self._validate_generated_project()
            structlogger.info(
                "project_generator.validation_success",
                attempts_left=attempts_left,
                max_retries=max_retries,
            )

            return commit_sha

        except ValidationError as e:
            structlogger.error(
                "project_generator.validation_error",
                error=str(e),
                attempts_left=attempts_left,
                max_retries=max_retries,
            )
            raise e

        except Exception as e:
            structlogger.error(
                "project_generator.generation_error",
                error=str(e),
                attempts_left=attempts_left,
                max_retries=max_retries,
            )
            raise e

    @observe(as_type="generation")
    async def generate_response(
        self,
        initial_messages: List[Dict[str, Any]],
        error_feedback_messages: List[Dict[str, Any]],
    ) -> Dict[str, Any]:
        """Generate Rasa assistant project data via the LLM."""
        try:
            async with self._get_client() as client:
                messages = [*initial_messages, *error_feedback_messages]
                response_format: Dict[str, Any] = {
                    "type": "json_schema",
                    "json_schema": {
                        "name": "rasa_project",
                        "schema": {
                            "type": "object",
                            "properties": {
                                "domain": self._domain_schema,
                                "flows": self._flows_schema,
                            },
                            "required": ["domain", "flows"],
                        },
                    },
                }
                response = await client.chat.completions.create(
                    model=config.OPENAI_MODEL,
                    messages=cast(Any, messages),
                    timeout=PROJECT_GENERATION_TIMEOUT,
                    temperature=config.OPENAI_TEMPERATURE,
                    response_format=cast(Any, response_format),
                )

            # Extract and store usage statistics
            usage_statistics = UsageStatistics.from_chat_completion_response(
                response,
                input_token_price=config.COPILOT_INPUT_TOKEN_PRICE,
                output_token_price=config.COPILOT_OUTPUT_TOKEN_PRICE,
                cached_token_price=config.COPILOT_CACHED_TOKEN_PRICE,
            )
            if usage_statistics:
                self._usage_statistics_attempts.append(usage_statistics)
                # Update the current Langfuse span with usage statistics
                PromptToBotLangfuseTelemetry.update_current_span_with_usage_statistics(
                    usage_statistics
                )

            content = response.choices[0].message.content
            if not content:
                error_message = "Empty response from LLM."
                structlogger.error(
                    "project_generator.generate_response.empty_llm_response",
                    event_info=error_message,
                    messages=messages,
                    response=response,
                )
                raise LLMGenerationError(error_message)

            try:
                return json.loads(content)
            except json.JSONDecodeError as e:
                error_message = f"Invalid JSON from LLM: {e}."
                structlogger.error(
                    "project_generator.generate_response.llm_response_invalid_json_format",
                    content=content,
                    event_info=error_message,
                    error=str(e),
                )
                raise LLMGenerationError(error_message)

        except openai.OpenAIError as e:
            error_message = f"OpenAI API error: {e}."
            structlogger.error(
                "project_generator.generate_response.openai_api_error",
                event_info=error_message,
                messages=messages,
                error=str(e),
            )
            raise LLMGenerationError(error_message)
        except asyncio.TimeoutError as e:
            error_message = "LLM request timed out."
            structlogger.error(
                "project_generator.generate_response.llm_request_timeout",
                event_info=error_message,
                messages=messages,
                error=str(e),
            )
            raise LLMGenerationError(error_message)

    @observe()
    async def _get_copilot_error_guidance(
        self,
        error: Exception,
        project_data: Dict[str, Any],
    ) -> Optional[str]:
        """Get copilot error analysis guidance for validation errors.

        Args:
            error: The validation error that occurred
            project_data: The project data that failed validation

        Returns:
            Copilot guidance text if available, None otherwise
        """
        try:
            # Build error message from exception
            logs = error.validation_logs if isinstance(error, ValidationError) else []
            error_message = str(error)
            if logs:
                error_message += "\n\n" + "\n".join(str(log) for log in logs)

            # Convert project_data to file content blocks for copilot analysis
            file_content_blocks: List[FileContent] = []
            # Add domain.yml
            domain_content = dump_obj_as_yaml_to_string(project_data["domain"])
            file_content_blocks.append(
                FileContent(
                    type="file", file_path="domain.yml", file_content=domain_content
                )
            )
            # Add flow files
            flows_dict = project_data.get("flows", {})
            for flow_id, flow_data in flows_dict.items():
                flow_file_path = self._path_for_flow(flow_id)
                single_flow_file_data = {"flows": {flow_id: flow_data}}
                flow_content = dump_obj_as_yaml_to_string(single_flow_file_data)
                file_content_blocks.append(
                    FileContent(
                        type="file", file_path=flow_file_path, file_content=flow_content
                    )
                )

            # Create message content blocks with log content and available files
            log_content_block = LogContent(
                type="log", content=error_message, context="validation_error"
            )
            context = CopilotContext(
                tracker_context=None,  # No conversation context needed
                copilot_chat_history=[
                    InternalCopilotRequestChatMessage(
                        role="internal_copilot_request",
                        content=[log_content_block, *file_content_blocks],
                        response_category=ResponseCategory.TRAINING_ERROR_LOG_ANALYSIS,
                    )
                ],
            )

            copilot_class = get_copilot_class()
            copilot = copilot_class()
            # Generate copilot response and handle it with the response handler.
            # Consume the stream to get the full response.
            (
                copilot_response_handler,
                generation_context,
            ) = await copilot.generate_response(context)
            async for _ in copilot_response_handler.stream():
                pass

            # Extract the full text from the handler
            full_text = copilot_response_handler.extract_text_from_generated_responses()
            return full_text if full_text else None

        except Exception as e:
            structlogger.warning(
                "project_generator.copilot_error_analysis_failed",
                error=str(e),
                event_info=(
                    "Failed to get copilot error guidance, continuing without it"
                ),
            )
            return None

    def _create_system_message(self) -> Dict[str, Any]:
        """Create a system message for skill generation."""
        system_prompt = self._skill_to_bot_system_prompt.render(
            flow_documentation_results=self._flow_documentation,
            domain_documentation_results=self._domain_documentation,
            custom_actions_documentation_results=self._custom_actions_documentation,
        )
        return {"role": ROLE_SYSTEM, "content": system_prompt}

    def _create_user_request_message(
        self, skill_description: str, project_data: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Create a user request message for skill generation."""
        user_request_prompt = self._skill_to_bot_user_prompt.render(
            skill_description=skill_description,
            project_data=project_data,
        )
        return {"role": ROLE_USER, "content": user_request_prompt}

    def _create_error_feedback_messages(
        self,
        project_data: Dict[str, Any],
        error: Exception,
        copilot_guidance: Optional[str] = None,
    ) -> List[Dict[str, Any]]:
        """Create error feedback messages for the LLM.

        Args:
            project_data: The project data that failed validation
            error: The validation error that occurred
            copilot_guidance: Optional copilot error analysis guidance
        """
        logs = error.validation_logs if isinstance(error, ValidationError) else []
        error_feedback_prompt = self._error_feedback_template.render(
            project_data=project_data,
            error=error,
            logs=logs,
            copilot_guidance=copilot_guidance,
        )
        return [
            {
                "role": ROLE_ASSISTANT,
                "content": json.dumps(project_data),
            },
            {
                "role": ROLE_USER,
                "content": error_feedback_prompt,
            },
        ]

    async def _validate_generated_project(self) -> None:
        """Validate the generated project using the validation service."""
        importer = self._create_importer()
        validation_error = await validate_project(importer)

        if validation_error:
            raise ValidationError(validation_error)

    def _get_endpoints_file(self) -> Path:
        """Get the endpoints file."""
        return self.project_folder / "endpoints.yml"

    def _get_config_file(self) -> Path:
        """Get the config file."""
        return self.project_folder / "config.yml"

    def get_training_input(self) -> TrainingInput:
        """Get the training input."""
        return TrainingInput(
            importer=self._create_importer(),
            endpoints_file=self._get_endpoints_file(),
            config_file=self._get_config_file(),
        )

    def _create_importer(self) -> TrainingDataImporter:
        """Create a training data importer from the current bot files."""
        try:
            if (self.project_folder / "domain.yml").exists():
                domain_path = self.project_folder / "domain.yml"
            else:
                domain_path = self.project_folder / "domain"

            return TrainingDataImporter.load_from_config(
                config_path=str(self._get_config_file()),
                domain_path=str(domain_path),
                training_data_paths=[str(self.project_folder / "data")],
                args={},
            )

        except Exception as e:
            raise ValidationError(f"Failed to create importer: {e}")

    def get_bot_files(
        self,
        allowed_file_extensions: Optional[List[str]] = None,
        exclude_docs_directory: bool = False,
        exclude_models_directory: bool = True,
    ) -> BotFiles:
        """Get the current bot files by reading from disk."""
        return get_bot_files(
            self.project_folder,
            allowed_file_extensions,
            exclude_docs_directory,
            exclude_models_directory,
        )

    def bot_file_paths(
        self, exclude_models_directory: bool = True
    ) -> Generator[Path, None, None]:
        """Get the paths of all bot files."""
        yield from bot_file_paths(self.project_folder, exclude_models_directory)

    def _get_bot_data_for_llm(self) -> Dict[str, Any]:
        """Get the current bot data for the LLM."""
        file_importer = self._create_importer()

        # only include data created by the user (or the builder llm)
        # avoid including to many defaults that are not customized
        domain = file_importer.get_user_domain()
        flows = file_importer.get_user_flows()

        return {
            "domain": domain.as_dict(should_clean_json=True),
            "flows": yaml_flows_io.get_flows_as_json(flows, should_clean_json=True),
        }

    def _path_for_flow(self, flow_id: str) -> str:
        """Get the path for a flow."""
        if flow_id.startswith("pattern_"):
            return f"data/patterns/{flow_id}.yml"
        else:
            return f"data/flows/{flow_id}.yml"

    async def _update_bot_files_from_llm_response(
        self, project_data: Dict[str, Any]
    ) -> str:
        """Update the bot files with generated data by writing to disk."""
        files: Dict[str, Optional[str]] = {
            "domain/domain.yml": dump_obj_as_yaml_to_string(project_data["domain"])
        }
        # split up flows into one file per flow in the /flows folder
        for flow_id, flow_data in project_data.get("flows", {}).items():
            flow_file_path = self._path_for_flow(flow_id)
            single_flow_file_data = {"flows": {flow_id: flow_data}}
            files[flow_file_path] = dump_obj_as_yaml_to_string(single_flow_file_data)

        # removes any other flows that the LLM didn't generate
        self._cleanup_flows()
        commit_sha = await self.update_bot_files(files)
        return commit_sha

    def _cleanup_flows(self) -> None:
        """Cleanup the flows folder."""
        flows_folder = self.project_folder / "data" / "flows"
        if flows_folder.exists():
            shutil.rmtree(flows_folder)
        flows_folder.mkdir(parents=True, exist_ok=True)

    async def update_bot_files(self, files: Dict[str, Optional[str]]) -> str:
        """Update bot files with new content by writing to disk."""
        # Acquire lock for entire operation (file writes + commit)
        async with self.git_service.git_operation():
            # Ensure git repository exists before committing
            self._ensure_git_repository()

            unsafe_write_to_bot_files(
                self.project_folder, files, fail_on_restricted_path=False
            )
            # Commit changes using internal method (lock already held)
            commit_sha = await self.git_service._commit_changes_internal(
                DEFAULT_COMMIT_INFO.model_copy(
                    update={"message": "Generated initial project files"}
                )
            )
            return commit_sha

    def ensure_all_files_are_writable(self, files: Dict[str, Optional[str]]) -> None:
        """Ensure all files are writable."""
        for filename, content in files.items():
            file_path = path_relative_to_project(self.project_folder, filename)
            if is_restricted_path(self.project_folder, file_path):
                raise InvalidPathException(
                    f"This file or folder is restricted from editing: {file_path}"
                )

    async def replace_all_bot_files(
        self, files: Dict[str, Optional[str]], commit_info: GitCommitInfo
    ) -> str:
        """Replace all bot files with new content, deleting files not in the request.

        Files/folders starting with .rasa/ or models/ are excluded from deletion.

        Args:
            files: Dictionary mapping file names to their content
            commit_info: info about the commit

        Returns:
            Commit SHA of the created commit
        """
        self.ensure_all_files_are_writable(files)

        # Validate file content state transitions before making any changes
        current_files = self.get_bot_files()
        validate_file_content_transitions(files, current_files, self.project_folder)

        # Acquire lock for entire operation
        async with self.git_service.git_operation():
            # Ensure git repository exists before committing
            self._ensure_git_repository()

            # Collect all existing files - any files not in the new `files` dict will be
            # deleted from this set
            existing_files = set(
                path.as_posix()
                for path in self.bot_file_paths(exclude_models_directory=True)
            )

            # Write all new files
            for filename, content in files.items():
                try:
                    file_path = path_relative_to_project(self.project_folder, filename)

                    # don't edit binary files
                    if content is None:
                        existing_files.discard(file_path.as_posix())
                        continue

                    unsafe_write_to_bot_files(
                        self.project_folder,
                        {filename: content},
                        fail_on_restricted_path=False,
                    )
                    existing_files.discard(file_path.as_posix())
                except Exception as e:
                    capture_exception_with_context(
                        e,
                        "project_generator.replace_all_bot_files.write_error",
                        extra={"file_path": filename},
                    )
                    continue

            # Delete files that weren't in the request
            for file_to_delete in existing_files:
                file_path = Path(file_to_delete)
                try:
                    file_path.unlink()
                except Exception as e:
                    capture_exception_with_context(
                        e,
                        "project_generator.replace_all_bot_files.delete_error",
                        extra={"file_path": file_path},
                    )

            # Clean up empty directories (except excluded ones)
            self._cleanup_empty_directories()
            # Commit changes to Git with AI-generated message
            return await self.unsafe_commit_changes(commit_info)

    def _cleanup_empty_directories(self) -> None:
        """Remove empty directories from the project folder.

        Excludes hidden files and directories, and models/ from cleanup.
        """
        # Walk directories in reverse order (deepest first)
        for dirpath, dirnames, filenames in os.walk(self.project_folder, topdown=False):
            # Skip if this is the project root
            if dirpath == str(self.project_folder):
                continue

            if is_restricted_path(self.project_folder, Path(dirpath)):
                continue

            relative_path = Path(dirpath).relative_to(self.project_folder)

            try:
                # Only remove if directory is empty
                if not os.listdir(dirpath):
                    os.rmdir(dirpath)
            except Exception as e:
                capture_exception_with_context(
                    e,
                    "project_generator.cleanup_empty_directories.error",
                    extra={"directory": relative_path.as_posix()},
                )

    def cleanup(self, skip_files: Optional[List[str]] = None) -> None:
        """Cleanup the project folder.

        Args:
            skip_files: List of file/directory names to skip during cleanup.
        """
        if skip_files is None:
            skip_files = []

        # Always include "lost+found" in skip files
        skip_files = list(skip_files) + ["lost+found"]

        # remove all the files and folders in the project folder resulting
        # in an empty folder
        for filename in os.listdir(self.project_folder):
            file_path = os.path.join(self.project_folder, filename)
            try:
                if filename in skip_files:
                    continue
                if os.path.isfile(file_path) or os.path.islink(file_path):
                    os.unlink(file_path)
                elif os.path.isdir(file_path):
                    shutil.rmtree(file_path)
            except Exception as e:
                structlogger.error(
                    "project_generator.cleanup_error",
                    error=str(e),
                    file_path=file_path,
                )

    def _ensure_git_repository(self) -> None:
        """Ensure the project folder is a Git repository on main branch."""
        # Initialize synchronously to avoid event loop conflicts in __init__
        if self.git_service.git_dir.exists():
            return

        structlogger.info(
            "project_generator.init_git_repository",
            project_folder=self.project_folder.as_posix(),
        )

        # Initialize Git repository synchronously
        self.git_service._initialize_git_repository()
        self.git_service._setup_git_configuration()
        self.git_service._create_gitignore()

        structlogger.info(
            "project_generator.git_repository_initialized",
            project_folder=self.project_folder.as_posix(),
        )

    def migrate_git_repository_if_needed(self) -> None:
        """Initialize Git and create an initial commit for existing projects.

        This migrates pre-existing builder instances that were created before
        Git integration existed. If the project folder contains files and no
        `.git` directory, we initialize a repo and create an initial commit
        capturing the current state. Hidden files and folders like `.rasa/` and
        large artifacts like `models/` are excluded via `.gitignore`.
        """
        try:
            if self.git_service.git_dir.exists():
                return

            # Only migrate if there is something to snapshot. Hidden files are
            # ignored by is_empty().
            if self.is_empty():
                return

            structlogger.info(
                "project_generator.migration_git_repository_start",
                project_folder=self.project_folder.as_posix(),
            )

            # Initialize repo and basic config + .gitignore (sync, safe in __init__)
            self.git_service._initialize_git_repository()
            self.git_service._setup_git_configuration()
            self.git_service._create_gitignore()

            # Stage and create an initial snapshot commit synchronously
            # It's fine if commit fails because nothing is staged.
            self.git_service.run_git_command_sync(["add", "."])
            try:
                self.git_service.run_git_command_sync(
                    [
                        "commit",
                        "-m",
                        "Initialize Git history from existing project state",
                    ]
                )
                structlogger.info(
                    "project_generator.migration_git_repository_completed",
                    project_folder=self.project_folder.as_posix(),
                )
            except subprocess.CalledProcessError as e:
                # No staged changes or commit failed; log and continue
                structlogger.warning(
                    "project_generator.migration_git_repository_commit_failed",
                    error=str(e),
                    project_folder=self.project_folder.as_posix(),
                )
        except Exception as e:
            structlogger.warning(
                "project_generator.migration_git_repository_failed",
                error=str(e),
                project_folder=self.project_folder.as_posix(),
            )

    async def unsafe_commit_changes(self, commit_info: GitCommitInfo) -> str:
        """Commit all changes.

        Args:
            commit_info: info about the commit

        Returns:
            Commit SHA of the created commit
        """
        try:
            # Ensure git repository exists (defensive in case migration didn't run)
            self._ensure_git_repository()

            # Generate commit message if not provided
            if commit_info.message is None:
                commit_info.message = await self._generate_commit_message()

            # Commit changes using internal method (lock already held)
            commit_sha = await self.git_service._commit_changes_internal(commit_info)

            return commit_sha

        except Exception as e:
            structlogger.error(
                "project_generator.git_commit_failed",
                error=str(e),
                project_folder=self.project_folder.as_posix(),
            )
            # Don't fail the operation if Git commit fails, return current commit
            return await self.git_service.get_current_commit_sha()

    @WelcomeMessageGenerationLangfuseTelemetry.trace_text_generation
    async def _generate_text(
        self, prompt: str, max_completion_tokens: int = 100
    ) -> ChatCompletion:
        """Generate simple text using OpenAI.

        Args:
            prompt: The text prompt to send to the model
            max_tokens: Maximum tokens to generate

        Returns:
            Chat Completion response

        Raises:
            LLMGenerationError: If generation fails
        """
        try:
            async with self._get_client() as client:
                response = await client.chat.completions.create(
                    model=config.OPENAI_MODEL,
                    messages=[{"role": "user", "content": prompt}],
                    temperature=0.1,  # Lower temperature for consistent messages
                    max_completion_tokens=max_completion_tokens,
                )

                if not response.choices[0].message.content:
                    raise LLMGenerationError("Empty response from LLM")

                return response

        except openai.OpenAIError as e:
            raise LLMGenerationError(f"OpenAI API error: {e}")
        except asyncio.TimeoutError:
            raise LLMGenerationError("LLM request timed out")

    @observe()
    async def _generate_commit_message(self) -> str:
        """Generate a meaningful commit message using AI based on the changes.

        Returns:
            A descriptive commit message based on the Git diff
        """
        try:
            # Get the diff of staged changes
            diff_output = (
                await self.git_service.run_git_command(
                    ["diff", "--name-status"], check_output=True
                )
                or ""
            )
            if not diff_output:
                return DEFAULT_COMMIT_MESSAGE

            # Get a more detailed diff for context (limited to avoid token limits)
            detailed_diff = (
                await self.git_service.run_git_command(
                    ["diff", "--unified=2"], check_output=True
                )
                or ""
            )

            # Limit the diff size to avoid token limits
            if detailed_diff and len(detailed_diff) > 2000:
                detailed_diff = detailed_diff[:2000] + "\n... (diff truncated)"

            # Render the prompt using the template
            prompt = self._commit_message_prompt_template.render(
                diff_output=diff_output,
                detailed_diff=detailed_diff,
                desired_length=config.COMMIT_MESSAGE_DESIRED_LENGTH,
            )

            # Update Langfuse span with input data
            CommitMessageGenerationLangfuseTelemetry.update_commit_message_generation_input(
                diff_output=diff_output,
                detailed_diff=detailed_diff,
                prompt=prompt,
            )

            # Use the existing LLM service to generate the commit message
            response = await self._generate_text(
                prompt, max_completion_tokens=config.COMMIT_MESSAGE_MAX_TOKENS
            )
            response_content = response.choices[0].message.content or ""

            # Clean up the response
            commit_message = response_content.strip().strip('"').strip("'")

            # Fallback to a reasonable default if generation fails or is too long
            if (
                not commit_message
                or len(commit_message) > config.COMMIT_MESSAGE_MAX_CHARACTERS
            ):
                commit_message = DEFAULT_COMMIT_MESSAGE

            # Update Langfuse span with output data
            CommitMessageGenerationLangfuseTelemetry.update_commit_message_generation_output(
                response_content=response_content,
                commit_message=commit_message,
            )

            return commit_message

        except Exception:
            structlogger.warning(
                "project_generator.commit_message_generation_failed",
                project_folder=self.project_folder.as_posix(),
            )
            # Fallback to generic message
            return DEFAULT_COMMIT_MESSAGE

    async def checkout_branch(
        self, branch_name: str, create_if_not_exists: bool = False
    ) -> str:
        """Checkout a Git branch.

        Args:
            branch_name: Name of the branch to checkout
            create_if_not_exists: Whether to create the branch if it doesn't exist

        Raises:
            GitOperationInProgressError: If another git operation is in progress
            subprocess.CalledProcessError: If the checkout fails
        """
        # Ensure repository exists before branch operations
        if not self.git_service.git_dir.exists():
            self._ensure_git_repository()
        await self.git_service.checkout_branch(branch_name, create_if_not_exists)
        return await self.git_service.get_current_commit_sha()

    async def _get_current_branch(self) -> str:
        """Get the current Git branch name."""
        return await self.git_service.get_current_branch()

    def _verify_bullet_points(self, response: str, max_amount: int) -> bool:
        """Verify that the response is in bullet point format."""
        lines = [line.strip() for line in response.splitlines() if line.strip()]
        if not lines:
            return False
        if len(lines) > max_amount:
            return False
        return all(BULLET_POINT_REGEX.match(line) for line in lines)

    @observe()
    async def generate_welcome_message(
        self, default_welcome_message: str, template_welcome_message: str
    ) -> str:
        """Generate a welcome message based on the generated flows.

        Returns:
            welcome message string
        """
        try:
            # Get generated flows
            flows = self._get_bot_data_for_llm().get("flows", {})
            if not flows:
                return default_welcome_message

            # Render the prompt using the template
            prompt = self._welcome_message_prompt_template.render(flows=flows)

            # Update Langfuse span with input data
            WelcomeMessageGenerationLangfuseTelemetry.update_welcome_message_generation_input(
                flows=flows,
                prompt=prompt,
            )

            # Use the existing LLM service to generate the example questions
            response = await self._generate_text(
                prompt, max_completion_tokens=config.WELCOME_MESSAGE_MAX_TOKENS
            )
            response_content = response.choices[0].message.content or ""

            # Initialize variables with defaults
            welcome_message = default_welcome_message

            # Clean up the response
            if (
                response_content
                and template_welcome_message
                and self._verify_bullet_points(response_content, 3)
            ):
                welcome_message = template_welcome_message.format(
                    example_questions=response_content
                )

            # Update Langfuse span with output data
            WelcomeMessageGenerationLangfuseTelemetry.update_welcome_message_generation_output(
                response_content=response_content,
                welcome_message=welcome_message,
            )

            return welcome_message

        except Exception:
            structlogger.warning(
                "project_generator.welcome_message_generation_failed",
                project_folder=self.project_folder.as_posix(),
            )
            # Fallback to generic message
            return default_welcome_message



from __future__ import annotations

import dataclasses
from pathlib import Path
from typing import Literal, Optional

import structlog

from rasa.cli.validation.config_path_validation import get_validated_path
from rasa.core.available_agents import DEFAULT_AGENTS_CONFIG_FOLDER, AvailableAgents
from rasa.core.config.available_endpoints import AvailableEndpoints
from rasa.core.config.credentials import CredentialsConfig
from rasa.core.config.message_procesing_config import MessageProcessingConfig
from rasa.shared.constants import (
    DEFAULT_CONFIG_PATH,
    DEFAULT_CREDENTIALS_PATH,
    DEFAULT_ENDPOINTS_PATH,
)

logger = structlog.get_logger(__name__)


@dataclasses.dataclass
class ValidatedConfigPaths:
    credentials_path: Optional[Path] = None
    endpoints_path: Optional[Path] = None
    message_processing_config_path: Optional[Path] = None


@dataclasses.dataclass
class ConfigPath:
    target: Optional[Path] = None
    type: Literal["credentials", "endpoints", "config"] = "endpoints"
    alternative_targets: list[Path] = dataclasses.field(default_factory=list)
    default_path: str = ""

    @staticmethod
    def _validate(
        config_path: ConfigPath, can_be_missing: bool = True
    ) -> Optional[Path]:
        validated_path = get_validated_path(
            config_path.target,
            config_path.type,
            config_path.alternative_targets,
            can_be_missing,
        )

        if validated_path is None and can_be_missing:
            logger.debug(
                f"No {config_path.type} configuration file found. "
                f"Proceeding without {config_path.type} configuration."
            )

        if isinstance(validated_path, str):
            validated_path = Path(validated_path)

        return validated_path


@dataclasses.dataclass
class CredentialsConfigPath(ConfigPath):
    def __init__(self, target: Optional[Path] = None):
        super().__init__(
            target=target,
            type="credentials",
            alternative_targets=[Path(DEFAULT_CREDENTIALS_PATH)],
            default_path=DEFAULT_CREDENTIALS_PATH,
        )

    @staticmethod
    def validate(
        target_config_path: Optional[Path] = None, can_be_missing: bool = True
    ) -> Optional[Path]:
        config_path = CredentialsConfigPath(target=target_config_path)
        return ConfigPath._validate(config_path, can_be_missing)

    @staticmethod
    def default_file_path() -> Path:
        return Path(DEFAULT_CREDENTIALS_PATH)

    @staticmethod
    def at_path(
        root_path: Path,
    ) -> Path:
        return root_path / CredentialsConfigPath.default_file_path()


@dataclasses.dataclass
class EndpointsConfigPath(ConfigPath):
    def __init__(self, target: Optional[Path] = None):
        super().__init__(
            target=target,
            type="endpoints",
            alternative_targets=[Path(DEFAULT_ENDPOINTS_PATH)],
            default_path=DEFAULT_ENDPOINTS_PATH,
        )

    @staticmethod
    def validate(
        target_config_path: Optional[Path] = None, can_be_missing: bool = True
    ) -> Optional[Path]:
        config_path = EndpointsConfigPath(target=target_config_path)
        return ConfigPath._validate(config_path, can_be_missing)

    @staticmethod
    def default_file_path() -> Path:
        return Path(DEFAULT_ENDPOINTS_PATH)

    @staticmethod
    def at_path(
        root_path: Path,
    ) -> Path:
        return root_path / EndpointsConfigPath.default_file_path()


@dataclasses.dataclass
class MessageProcessingConfigPath(ConfigPath):
    def __init__(self, target: Optional[Path] = None):
        super().__init__(
            target=target,
            type="config",
            alternative_targets=[Path(DEFAULT_CONFIG_PATH)],
            default_path=DEFAULT_CONFIG_PATH,
        )

    @staticmethod
    def validate(
        target_config_path: Optional[Path] = None, can_be_missing: bool = True
    ) -> Optional[Path]:
        config_path = MessageProcessingConfigPath(target=target_config_path)
        return ConfigPath._validate(config_path, can_be_missing)

    @staticmethod
    def default_file_path() -> Path:
        return Path(DEFAULT_CONFIG_PATH)

    @staticmethod
    def at_path(
        root_path: Path,
    ) -> Path:
        return root_path / MessageProcessingConfigPath.default_file_path()


class Configuration:
    _instance: Optional[Configuration] = None

    def __init__(
        self,
        endpoints: AvailableEndpoints,
        available_agents: AvailableAgents,
        credentials: Optional[CredentialsConfig] = None,
        message_processing_config: Optional[MessageProcessingConfig] = None,
    ):
        self.credentials = credentials
        self.available_agents = available_agents
        self.endpoints = endpoints
        self.message_processing_config = message_processing_config

    @classmethod
    def initialise_empty(cls) -> Configuration:
        if cls._instance is None:
            cls._instance = Configuration(
                credentials=None,
                endpoints=AvailableEndpoints(),
                available_agents=AvailableAgents(),
                message_processing_config=None,
            )
        return cls._instance

    @classmethod
    def initialise_endpoints(cls, endpoints_path: Path) -> Configuration:
        logger.debug(
            "configuration.initialise_endpoints.start", endpoints_path=endpoints_path
        )

        validated_endpoints_path = EndpointsConfigPath.validate(endpoints_path)
        endpoints = (
            AvailableEndpoints.read_endpoints(validated_endpoints_path)
            if validated_endpoints_path
            else AvailableEndpoints()
        )
        if cls._instance is None:
            cls._instance = Configuration(
                endpoints=endpoints,
                available_agents=AvailableAgents(),
                credentials=None,
                message_processing_config=None,
            )
        else:
            cls._instance.endpoints = endpoints
        return cls._instance

    @classmethod
    def initialise_empty_endpoints(cls) -> Configuration:
        logger.debug("configuration.initialise_empty_endpoints.start")
        if cls._instance is None:
            cls._instance = Configuration(
                endpoints=AvailableEndpoints(),
                available_agents=AvailableAgents(),
                credentials=None,
                message_processing_config=None,
            )
        else:
            cls._instance.endpoints = AvailableEndpoints()
        return cls._instance

    @classmethod
    def initialise_message_processing(
        cls, message_processing_config_path: Path
    ) -> Configuration:
        logger.debug(
            "configuration.initialise_message_processing.start",
            message_processing_config_path=message_processing_config_path,
        )
        validated_message_processing_path = MessageProcessingConfigPath.validate(
            message_processing_config_path
        )
        message_processing_config = (
            MessageProcessingConfig.load_from_file(validated_message_processing_path)
            if validated_message_processing_path
            else None
        )
        if cls._instance is None:
            cls._instance = Configuration(
                message_processing_config=message_processing_config,
                credentials=None,
                endpoints=AvailableEndpoints(),
                available_agents=AvailableAgents(),
            )
        else:
            cls._instance.message_processing_config = message_processing_config
        return cls._instance

    @classmethod
    def initialise_credentials(cls, credentials_path: Path) -> Configuration:
        logger.debug(
            "configuration.initialise_credentials.start",
            credentials_path=credentials_path,
        )

        validated_credentials_path = CredentialsConfigPath.validate(credentials_path)

        credentials = (
            CredentialsConfig.load_from_file(validated_credentials_path)
            if validated_credentials_path
            else None
        )
        if cls._instance is None:
            cls._instance = Configuration(
                credentials=credentials,
                endpoints=AvailableEndpoints(),
                available_agents=AvailableAgents(),
                message_processing_config=None,
            )
        else:
            cls._instance.credentials = credentials
        return cls._instance

    @classmethod
    def initialise_sub_agents(cls, sub_agents_path: Optional[Path]) -> Configuration:
        sub_agents_folder = (
            str(sub_agents_path)
            if sub_agents_path is not None
            else DEFAULT_AGENTS_CONFIG_FOLDER
        )
        logger.debug(
            "configuration.initialise_sub_agents.start",
            sub_agents_folder=sub_agents_folder,
        )
        available_agents = AvailableAgents.read_from_folder(sub_agents_folder)

        if cls._instance is None:
            cls._instance = Configuration(
                endpoints=AvailableEndpoints(),
                available_agents=available_agents,
                credentials=None,
                message_processing_config=None,
            )
        else:
            cls._instance.available_agents = available_agents
        return cls._instance

    @classmethod
    def get_instance(cls) -> Configuration:
        if cls._instance is None:
            raise Exception(
                "Configuration not initialized. "
                "Call appropriate 'initialise' methods to "
                "load the config when Rasa Pro starts: "
                "initialise_endpoints(), "
                "initialise_sub_agents(), "
                "initialise_credentials(), "
                "initialise_message_processing()"
            )
        return cls._instance



from collections import defaultdict
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Union

import rich

import rasa.shared.data
from rasa.dialogue_understanding_test.command_metrics import CommandMetrics
from rasa.dialogue_understanding_test.constants import SCHEMA_FILE_PATH
from rasa.dialogue_understanding_test.du_test_case import (
    KEY_CHOICES,
    KEY_COMPLETION_TOKENS,
    KEY_PROMPT_TOKENS,
)
from rasa.dialogue_understanding_test.du_test_result import (
    KEY_COMMANDS_F1_MACRO,
    KEY_COMMANDS_F1_MICRO,
    KEY_COMMANDS_F1_WEIGHTED,
    DialogueUnderstandingTestSuiteResult,
    FailedTestStep,
)
from rasa.e2e_test.constants import KEY_TEST_CASE, KEY_TEST_CASES
from rasa.e2e_test.e2e_test_case import (
    DialogueUnderstandingTestCase,
    Fixture,
    Metadata,
    TestSuite,
)
from rasa.e2e_test.stub_custom_action import StubCustomAction
from rasa.e2e_test.utils.io import (
    check_beta_feature_flag_for_custom_actions_stubs,
    extract_fixtures,
    extract_metadata,
    extract_stub_custom_actions,
    extract_test_case_from_path,
    is_test_case_file,
    parse_raw_yaml,
    validate_path_to_test_cases,
    validate_test_case,
)
from rasa.shared.core.flows import FlowsList
from rasa.shared.exceptions import DuplicateFixtureException, RasaException
from rasa.shared.nlu.constants import (
    KEY_LATENCY,
    KEY_PROMPT_NAME,
    KEY_SYSTEM_PROMPT,
    KEY_USER_PROMPT,
)
from rasa.shared.utils.yaml import (
    read_schema_file,
    validate_yaml_content_using_schema,
    write_yaml,
)

if TYPE_CHECKING:
    from rasa.dialogue_understanding.commands import Command


def read_du_test_schema() -> Union[List[Any], Dict[str, Any]]:
    """Read the schema for the dialogue understanding test cases.

    Returns:
        Dict[str, Any]: Schema for the dialogue understanding test cases.
    """
    return read_schema_file(SCHEMA_FILE_PATH)


def read_test_suite(
    test_case_path: str,
    flows: FlowsList,
    custom_command_classes: List["Command"] = [],
    remove_default_commands: List[str] = [],
) -> TestSuite:
    """Read the test cases from the given test case path.

    Args:
        test_case_path: Path to the test cases.
        flows: List of flows.
        custom_command_classes: Custom command classes to use in the test cases.
        remove_default_commands: Default commands to remove from the test cases.

    Returns:
        TestSuite: Test suite containing the dialogue understanding test cases.
    """

    def _extract_test_cases(
        test_file_content: dict,
        test_case_name: str,
        test_file: str,
        flows: FlowsList,
        custom_command_classes: List["Command"] = [],
        remove_default_commands: List[str] = [],
    ) -> List[DialogueUnderstandingTestCase]:
        """Extract test cases from the test file content.

        Args:
            test_file_content: Content of the test file.
            test_case_name: Name of the test case to extract.
            test_file: Path to the test file.
            flows: List of flows.

        Returns:
            List of test cases.
        """
        test_cases_content = test_file_content.get(KEY_TEST_CASES) or []
        if test_case_name:
            return [
                DialogueUnderstandingTestCase.from_dict(
                    test_case_dict,
                    flows,
                    file=test_file,
                    custom_command_classes=custom_command_classes,
                    remove_default_commands=remove_default_commands,
                )
                for test_case_dict in test_cases_content
                if test_case_name == test_case_dict.get(KEY_TEST_CASE)
            ]
        return [
            DialogueUnderstandingTestCase.from_dict(
                test_case_dict,
                flows,
                file=test_file,
                custom_command_classes=custom_command_classes,
                remove_default_commands=remove_default_commands,
            )
            for test_case_dict in test_cases_content
        ]

    # Extract test case path and name
    test_case_path, test_case_name = extract_test_case_from_path(test_case_path)
    validate_path_to_test_cases(test_case_path)

    # Load test files and schema
    test_files = rasa.shared.data.get_data_files([test_case_path], is_test_case_file)
    test_schema = read_du_test_schema()

    # Initialize containers
    input_test_cases = []
    fixtures: Dict[str, Fixture] = {}
    metadata: Dict[str, Metadata] = {}
    stub_custom_actions: Dict[str, StubCustomAction] = {}
    fixture_errors: List[str] = []

    # Process each test file
    for test_file in test_files:
        test_file_content = parse_raw_yaml(Path(test_file).read_text(encoding="utf-8"))

        # Validate YAML content using the provided function
        validate_yaml_content_using_schema(test_file_content, test_schema)

        # Parse test cases, fixtures, metadata, and stub custom actions
        test_cases = _extract_test_cases(
            test_file_content,
            test_case_name,
            test_file,
            flows,
            custom_command_classes,
            remove_default_commands,
        )
        try:
            fixtures.update(extract_fixtures(test_file_content, fixtures, test_file))
        except DuplicateFixtureException as e:
            fixture_errors.append(str(e))
        metadata.update(extract_metadata(test_file_content, metadata))
        stub_custom_actions.update(
            extract_stub_custom_actions(test_file_content, test_file)
        )
        input_test_cases.extend(test_cases)

    # Report all fixture errors at once
    if fixture_errors:
        raise RasaException(
            "Duplicate fixtures found in file(s):\n  - " + "\n  - ".join(fixture_errors)
        )

    validate_test_case(test_case_name, input_test_cases, fixtures, metadata)
    if stub_custom_actions:
        check_beta_feature_flag_for_custom_actions_stubs()

    return TestSuite(
        input_test_cases,
        list(fixtures.values()),
        list(metadata.values()),
        stub_custom_actions,
    )


def write_test_results_to_file(
    test_suite_result: DialogueUnderstandingTestSuiteResult,
    output_file: str,
    output_prompt: bool,
) -> None:
    """Write the test results to the given output file.

    Args:
        test_suite_result: Test results suite containing the test results.
        output_file: Path to the output file.
        output_prompt: Whether to log the prompt or not.
    """
    write_yaml(
        test_suite_result.to_dict(output_prompt),
        target=output_file,
        should_preserve_key_order=True,
    )

    print(f"Results written to '{output_file}'.")


def print_test_results(
    test_suite_result: DialogueUnderstandingTestSuiteResult,
    output_prompt: bool,
) -> None:
    """Print the result of the test run.

    Example output (truncated for brevity):
        ====== FAILURES ======

        ---------------- test_case: rasa-calm-demo/dialogue_understanding_tests/
        immediate_cancellation_and_start_of_new_flow.yml::user immediately cancels
        and starts new flow -----------------

        == failure starting at user message 'I want to send money'.

        -- COMMAND GENERATOR(s) --
        SingleStepLLMCommandGenerator

        -- CONVERSATION --
        user: I want to send money
        ---EXPECTED---                               | ---PREDICTED---
        StartFlow(transfer_money)                    | StartFlow(transfer_money)
        SetSlot(transfer_money_amount_of_money, 878) |

        ...
        ====== COMMAND METRICS ======
        set slot (2 commands in total):
          tp: 0 fp: 1 fn: 2
          precision: 0.0000
          recall   : 0.0000
          f1       : 0.0000

        cancel flow (1 commands in total):
          tp: 1 fp: 0 fn: 0
          precision: 1.0000
          recall   : 1.0000
          f1       : 1.0000

        ...
        ====== LATENCY METRICS ======
        p50: 0.00065571
        p90: 0.00074687
        p99: 0.00077837
        ====== PROMPT TOKEN METRICS ======
        p50: 1336.00
        p90: 1389.50
        p99: 1401.65
        ====== COMPLETION TOKEN METRICS ======
        p50: 12.00
        p90: 15.80
        p99: 16.88

        ====== 1 failed test cases, 0 passed test cases ======
        ====== 2 failed user steps, 1 passed user steps (accuracy: 0.3333) ======

    Args:
        test_suite_result: Test results suite containing the test results.
        output_prompt: Whether to log the prompt or not.
    """
    if (
        test_suite_result.number_of_passed_tests
        + test_suite_result.number_of_failed_tests
        == 0
    ):
        # no tests were run, print error
        rasa.shared.utils.cli.print_error(
            rasa.shared.utils.cli.pad("No test cases found.", char="!")
        )
        return

    if test_suite_result.number_of_failed_tests > 0:
        # print failure headline
        print()
        rich.print(
            f"[bold][red3]"
            f"{rasa.shared.utils.cli.pad('FAILURES', char='=')}"
            f"[/red3][/bold]"
        )

        # print failed test steps
        print_failed_cases(test_suite_result, output_prompt=output_prompt)

    print_f1_summary(test_suite_result)
    print_command_summary(test_suite_result.command_metrics)
    print_latency_and_token_metrics(test_suite_result)
    print_final_line(test_suite_result)


def print_failed_cases(
    test_suite_result: DialogueUnderstandingTestSuiteResult,
    output_prompt: bool,
) -> None:
    """Print the details of a failed test case."""
    # Group the failed test steps by test case
    step_groups = defaultdict(list)
    for step in test_suite_result.failed_test_steps:
        key = f"{step.file}::{step.test_case_name}"
        step_groups[key].append(step)

    for failed_test_case, failed_test_steps in step_groups.items():
        fail_headline = f"test_case: {failed_test_case}"
        print()
        rasa.shared.utils.cli.print_error(
            f"{rasa.shared.utils.cli.pad(fail_headline, char='-')}\n"
        )
        print(f"Number of failed steps: {len(failed_test_steps)}")
        for step in failed_test_steps:
            print()
            rasa.shared.utils.cli.print_info(
                f"== failure starting at user message "
                f"'{step.failed_user_utterance}'. "
            )

            rich.print("\n[red3]-- COMMAND GENERATOR(s) --[/red3]")
            rich.print("\n".join(step.command_generators))
            if output_prompt:
                print_prompt(step)
            rich.print("\n[red3]-- CONVERSATION --[/red3]")
            rich.print("\n".join(step.conversation_with_diff))
            print_llm_output(step)


def print_prompt(step: FailedTestStep) -> None:
    if step.prompts is None:
        return
    prompts = step.prompts

    rich.print("\n[red3]-- PROMPT(s) --[/red3]")
    for component, component_prompts in prompts.items():
        rich.print(f"[bold]{component}[/bold]")
        for prompt_data in component_prompts:
            rich.print(
                f"[bold]  prompt name      [/bold]: {prompt_data[KEY_PROMPT_NAME]}"
            )
            rich.print(
                f"[bold]  prompt tokens    [/bold]: {prompt_data[KEY_PROMPT_TOKENS]}"
            )
            rich.print(
                f"[bold]  completion tokens[/bold]: "
                f"{prompt_data[KEY_COMPLETION_TOKENS]}"
            )
            rich.print(f"[bold]  latency          [/bold]: {prompt_data[KEY_LATENCY]}")
            if KEY_SYSTEM_PROMPT in prompt_data:
                rich.print(
                    f"[bold]  system prompt    [/bold]: "
                    f"{prompt_data[KEY_SYSTEM_PROMPT]}"
                )
            rich.print(
                f"[bold]  user prompt      [/bold]: {prompt_data[KEY_USER_PROMPT]}"
            )


def print_llm_output(step: FailedTestStep) -> None:
    if not step.prompts:
        return

    for component, component_prompts in step.prompts.items():
        for prompt_data in component_prompts:
            if KEY_CHOICES in prompt_data:
                rich.print(f"\n[red3]-- LLM ouptut for {component} --[/red3]")
                rich.print(prompt_data.get(KEY_CHOICES))
                rich.print("[red3]-------------[/red3]")


def print_f1_summary(result: DialogueUnderstandingTestSuiteResult) -> None:
    """Print the f1 summary."""
    print()
    rasa.shared.utils.cli.print_info(rasa.shared.utils.cli.pad("COMMANDS F1"))
    rasa.shared.utils.cli.print_info(
        f"macro           : {result.f1_score[KEY_COMMANDS_F1_MACRO]:.8f}"
    )
    rasa.shared.utils.cli.print_info(
        f"micro           : {result.f1_score[KEY_COMMANDS_F1_MICRO]:.8f}"
    )
    rasa.shared.utils.cli.print_info(
        f"weighted average: {result.f1_score[KEY_COMMANDS_F1_WEIGHTED]:.8f}"
    )


def print_command_summary(metrics: Dict[str, CommandMetrics]) -> None:
    """Print the command summary.

    Args:
        metrics: Dict of command to precision, recall, f1, etc. scores
    """
    print()
    rasa.shared.utils.cli.print_info(rasa.shared.utils.cli.pad("COMMAND METRICS"))

    for command_name, command_metric in metrics.items():
        rasa.shared.utils.cli.print_info(
            f"{command_name} " f"({command_metric.total_count} commands in total):"
        )
        rasa.shared.utils.cli.print_info(
            f"  tp: {command_metric.tp} "
            f"fp: {command_metric.fp} "
            f"fn: {command_metric.fn}"
        )
        rasa.shared.utils.cli.print_info(
            f"  precision: {command_metric.get_precision():.4f}"
        )
        rasa.shared.utils.cli.print_info(
            f"  recall   : {command_metric.get_recall():.4f}"
        )
        rasa.shared.utils.cli.print_info(
            f"  f1       : {command_metric.get_f1_score():.4f}"
        )


def print_latency_and_token_metrics(
    result: DialogueUnderstandingTestSuiteResult,
) -> None:
    """Print the latency and token metrics."""
    print()
    rasa.shared.utils.cli.print_info(rasa.shared.utils.cli.pad("LATENCY METRICS"))
    for component, latency_metric in result.latency_metrics.items():
        rasa.shared.utils.cli.print_info(f"--- {component} ---")
        for key, value in latency_metric.items():
            rasa.shared.utils.cli.print_info(f"{key}: {value:.8f}")

    rasa.shared.utils.cli.print_info(rasa.shared.utils.cli.pad("PROMPT TOKEN METRICS"))
    for component, prompt_token_metric in result.prompt_token_metrics.items():
        rasa.shared.utils.cli.print_info(f"--- {component} ---")
        for key, value in prompt_token_metric.items():
            rasa.shared.utils.cli.print_info(f"{key}: {value:.2f}")

    rasa.shared.utils.cli.print_info(
        rasa.shared.utils.cli.pad("COMPLETION TOKEN METRICS")
    )
    for component, completion_token_metric in result.completion_token_metrics.items():
        rasa.shared.utils.cli.print_info(f"--- {component} ---")
        for key, value in completion_token_metric.items():
            rasa.shared.utils.cli.print_info(f"{key}: {value:.2f}")


def print_final_line(test_suite_result: DialogueUnderstandingTestSuiteResult) -> None:
    """Print the final line of the test output.

    Args:
        test_suite_result: Test suite result.
    """
    has_failed = test_suite_result.number_of_failed_tests > 0

    final_line_color = "green3" if not has_failed else "red3"

    def _print_accuracy_line(number_passed: int, number_failed: int, type: str) -> None:
        if number_passed > 0:
            line = rasa.shared.utils.cli.pad(
                f"[bold red3]{number_failed} failed {type}[/bold red3]"
                f"[bright_white], [/bright_white]"
                f"[bold green3]{number_passed} passed {type}[/bold green3]"
                f"[bright_white] "
                f"(accuracy:  {number_passed / (number_passed + number_failed):.4f})"
                f"[/bright_white]"
            )
        else:
            line = rasa.shared.utils.cli.pad(
                f"[bold red3]{number_failed} failed {type}[/bold red3]"
                f"[bright_white], [/bright_white]"
                f"[bold green3]{number_passed} passed {type}[/bold green3]"
            )
        rich.print(f"[{final_line_color}]{line}[/{final_line_color}]")

    _print_accuracy_line(
        test_suite_result.number_of_passed_tests,
        test_suite_result.number_of_failed_tests,
        "test cases",
    )

    _print_accuracy_line(
        test_suite_result.number_of_passed_user_utterances,
        test_suite_result.number_of_failed_user_utterances,
        "user steps",
    )



import dataclasses
import glob
import importlib.resources
import json
import os.path
import re
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Text, Tuple

import dotenv
import structlog
from jinja2 import Template
from pydantic import ValidationError

import rasa.shared.utils.io
from rasa.core.config.available_endpoints import AvailableEndpoints
from rasa.core.constants import (
    UTTER_SOURCE_METADATA_KEY,
)
from rasa.core.information_retrieval import (
    InformationRetrieval,
    InformationRetrievalException,
    SearchResult,
    SearchResultList,
    create_from_endpoint_config,
)
from rasa.core.information_retrieval.faiss import FAISS_Store
from rasa.core.policies.enterprise_search_policy_config import (
    DEFAULT_EMBEDDINGS_CONFIG,
    DEFAULT_ENTERPRISE_SEARCH_CONFIG,
    DEFAULT_LLM_CONFIG,
    DEFAULT_VECTOR_STORE_TYPE,
    SOURCE_PROPERTY,
    EnterpriseSearchPolicyConfig,
)
from rasa.core.policies.policy import Policy, PolicyPrediction
from rasa.dialogue_understanding.generator.constants import (
    LLM_CONFIG_KEY,
)
from rasa.dialogue_understanding.patterns.cannot_handle import (
    CannotHandlePatternFlowStackFrame,
)
from rasa.dialogue_understanding.patterns.internal_error import (
    InternalErrorPatternFlowStackFrame,
)
from rasa.dialogue_understanding.stack.frames import (
    DialogueStackFrame,
    PatternFlowStackFrame,
    SearchStackFrame,
)
from rasa.engine.graph import ExecutionContext
from rasa.engine.recipes.default_recipe import DefaultV1Recipe
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.exceptions import EnterpriseSearchPolicyError
from rasa.graph_components.providers.forms_provider import Forms
from rasa.graph_components.providers.responses_provider import Responses
from rasa.shared.constants import (
    EMBEDDINGS_CONFIG_KEY,
    MODEL_CONFIG_KEY,
    MODEL_GROUP_ID_CONFIG_KEY,
    MODEL_NAME_CONFIG_KEY,
    PROVIDER_CONFIG_KEY,
    RASA_PATTERN_CANNOT_HANDLE_NO_RELEVANT_ANSWER,
)
from rasa.shared.core.constants import (
    ACTION_CANCEL_FLOW,
    ACTION_METADATA_MESSAGE_KEY,
    ACTION_METADATA_TEXT_KEY,
    ACTION_SEND_TEXT_NAME,
    DEFAULT_SLOT_NAMES,
    MOCKED_DATETIME_SLOT,
)
from rasa.shared.core.domain import Domain
from rasa.shared.core.events import BotUttered, Event, UserUttered
from rasa.shared.core.generator import TrackerWithCachedStates
from rasa.shared.core.trackers import DialogueStateTracker, EventVerbosity
from rasa.shared.exceptions import FileIOException, RasaException
from rasa.shared.nlu.constants import (
    KEY_COMPONENT_NAME,
    KEY_LLM_RESPONSE_METADATA,
    KEY_PROMPT_NAME,
    KEY_USER_PROMPT,
    PROMPTS,
)
from rasa.shared.nlu.training_data.training_data import TrainingData
from rasa.shared.providers.embedding._langchain_embedding_client_adapter import (
    _LangchainEmbeddingClientAdapter,
    embedding_metadata_context,
)
from rasa.shared.providers.llm.llm_response import LLMResponse, measure_llm_latency
from rasa.shared.utils.constants import (
    LANGFUSE_METADATA_AGENT_ID,
    LANGFUSE_METADATA_COMPONENT_NAME,
    LANGFUSE_METADATA_CUSTOM_METADATA,
    LANGFUSE_METADATA_MODEL_ID,
    LANGFUSE_METADATA_SESSION_ID,
    LANGFUSE_METADATA_TAGS,
    LOG_COMPONENT_SOURCE_METHOD_FINGERPRINT_ADDON,
    LOG_COMPONENT_SOURCE_METHOD_INIT,
)
from rasa.shared.utils.datetime_utils import resolve_datetime
from rasa.shared.utils.health_check.embeddings_health_check_mixin import (
    EmbeddingsHealthCheckMixin,
)
from rasa.shared.utils.health_check.llm_health_check_mixin import LLMHealthCheckMixin
from rasa.shared.utils.io import deep_container_fingerprint
from rasa.shared.utils.llm import (
    LLMInput,
    acompletion_with_streaming,
    embedder_factory,
    get_prompt_template,
    llm_factory,
    sanitize_message_for_prompt,
    tracker_as_readable_transcript,
)
from rasa.telemetry import (
    track_enterprise_search_policy_predict,
    track_enterprise_search_policy_train_completed,
    track_enterprise_search_policy_train_started,
)

if TYPE_CHECKING:
    from langchain.schema.embeddings import Embeddings

    from rasa.core.channels.channel import OutputChannel
    from rasa.core.featurizers.tracker_featurizers import TrackerFeaturizer

from rasa.utils.log_utils import log_llm

structlogger = structlog.get_logger()

dotenv.load_dotenv("./.env")

ENTERPRISE_SEARCH_PROMPT_FILE_NAME = "enterprise_search_policy_prompt.jinja2"
ENTERPRISE_SEARCH_CONFIG_FILE_NAME = "config.json"

SEARCH_RESULTS_METADATA_KEY = "search_results"
SEARCH_QUERY_METADATA_KEY = "search_query"

DEFAULT_ENTERPRISE_SEARCH_PROMPT_TEMPLATE = importlib.resources.read_text(
    "rasa.core.policies", "enterprise_search_prompt_template.jinja2"
)

DEFAULT_ENTERPRISE_SEARCH_PROMPT_WITH_CITATION_TEMPLATE = importlib.resources.read_text(
    "rasa.core.policies", "enterprise_search_prompt_with_citation_template.jinja2"
)

DEFAULT_ENTERPRISE_SEARCH_PROMPT_WITH_RELEVANCY_CHECK_AND_CITATION_TEMPLATE = (
    importlib.resources.read_text(
        "rasa.core.policies",
        "enterprise_search_prompt_with_relevancy_check_and_citation_template.jinja2",
    )
)

_ENTERPRISE_SEARCH_ANSWER_NOT_RELEVANT_PATTERN = re.compile(r"\[NO_RAG_ANSWER\]")

_ENTERPRISE_SEARCH_CITATION_PATTERN = re.compile(r"\[([^\]]+)\]")


class VectorStoreConnectionError(RasaException):
    """Exception raised for errors in connecting to the vector store."""


class VectorStoreConfigurationError(RasaException):
    """Exception raised for errors in vector store configuration."""


@dataclasses.dataclass
class _RelevancyCheckResponse:
    answer: Optional[str]
    relevant: bool


@DefaultV1Recipe.register(
    DefaultV1Recipe.ComponentType.POLICY_WITH_END_TO_END_SUPPORT, is_trainable=True
)
class EnterpriseSearchPolicy(LLMHealthCheckMixin, EmbeddingsHealthCheckMixin, Policy):
    """Policy which uses a vector store and LLMs to respond to user messages.

    The policy uses a vector store and LLMs to respond to user messages. The
    vector store is used to retrieve the most relevant responses to the user
    message. The LLMs are used to rank the responses and select the best
    response. The policy can be used to respond to user messages without
    training data.

    Example Configuration:

        policies:
            # - ...
            - name: EnterpriseSearchPolicy
              vector_store:
                type: "milvus"
                <vector_store_config>
            # - ...
    """

    @staticmethod
    def does_support_stack_frame(frame: DialogueStackFrame) -> bool:
        """Checks if the policy supports the given stack frame."""
        return isinstance(frame, SearchStackFrame)

    @staticmethod
    def get_default_config() -> Dict[str, Any]:
        """Returns the default config of the policy."""
        return DEFAULT_ENTERPRISE_SEARCH_CONFIG

    def __init__(
        self,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
        vector_store: Optional[InformationRetrieval] = None,
        featurizer: Optional["TrackerFeaturizer"] = None,
        prompt_template: Optional[Text] = None,
    ) -> None:
        """Constructs a new Policy object."""
        super().__init__(config, model_storage, resource, execution_context, featurizer)

        parsed_config = EnterpriseSearchPolicyConfig.from_dict(config)

        # Vector store object and configuration
        self.vector_store = vector_store
        self.vector_store_config = parsed_config.vector_store_config
        self.vector_search_threshold = parsed_config.vector_store_threshold
        self.vector_store_type = parsed_config.vector_store_type

        # Resolved embeddings configuration for encoding the search query
        self.embeddings_config = parsed_config.embeddings_config

        # Resolved LLM Configuration for response generation
        self.llm_config = parsed_config.llm_config

        # Maximum number of turns to include in the prompt
        self.max_history = parsed_config.max_history

        # Maximum number of messages to include in the search query
        self.max_messages_in_query = parsed_config.max_messages_in_query

        # Boolean to enable/disable tracing of prompt tokens
        self.trace_prompt_tokens = parsed_config.trace_prompt_tokens

        # Boolean to enable/disable the use of LLM for response generation
        self.use_llm = parsed_config.use_generative_llm

        # Boolean to enable/disable citation generation. This flag enables citation
        # logic, but it only takes effect if `use_llm` is True.
        self.citation_enabled = parsed_config.enable_citation

        # Boolean to enable/disable the use of relevancy check alongside answer
        # generation. This flag enables citation logic, but it only takes effect if
        # `use_llm` is True.
        self.relevancy_check_enabled = parsed_config.check_relevancy

        # Datetime configuration
        self.include_date_time = parsed_config.include_date_time
        self.timezone = parsed_config.timezone

        # Resolve the prompt template. The prompt will only be used if the 'use_llm' is
        # set to True.
        self.prompt_template = prompt_template or get_prompt_template(
            jinja_file_path=parsed_config.prompt_template,
            default_prompt_template=self._select_default_prompt_template_based_on_features(
                parsed_config.check_relevancy, parsed_config.enable_citation
            ),
            log_source_component=EnterpriseSearchPolicy.__name__,
            log_source_method=LOG_COMPONENT_SOURCE_METHOD_INIT,
        )

    @classmethod
    def _create_plain_embedder(cls, embeddings_config: Dict[Text, Any]) -> "Embeddings":
        """Creates an embedder based on the given configuration.

        Args:
            embeddings_config: A resolved embeddings configuration. Resolved means the
            configuration is either:
                - A reference to a model group that has already been expanded into
                  its corresponding configuration using the information from
                  `endpoints.yml`, or
                - A full configuration for the embedder defined directly (i.e. not
                  relying on model groups or indirections).

        Returns:
            The embedder.
        """
        # Copy the config so original config is not modified
        embeddings_config = embeddings_config.copy()
        client = embedder_factory(embeddings_config, DEFAULT_EMBEDDINGS_CONFIG)
        # Wrap the embedding client in the adapter
        return _LangchainEmbeddingClientAdapter(client)

    @classmethod
    def _add_prompt_and_llm_response_to_latest_message(
        cls,
        tracker: DialogueStateTracker,
        prompt_name: str,
        user_prompt: str,
        llm_response: Optional[LLMResponse] = None,
    ) -> None:
        """Stores the prompt and LLMResponse metadata in the tracker.

        Args:
            tracker: The DialogueStateTracker containing the current conversation state.
            prompt_name: A name identifying prompt usage.
            user_prompt: The user prompt that was sent to the LLM.
            llm_response: The response object from the LLM (None if no response).
        """
        from rasa.dialogue_understanding.utils import record_commands_and_prompts

        if not record_commands_and_prompts:
            return

        if not tracker.latest_message:
            return

        parse_data = tracker.latest_message.parse_data
        if PROMPTS not in parse_data:
            parse_data[PROMPTS] = []  # type: ignore[literal-required]

        prompt_data: Dict[Text, Any] = {
            KEY_COMPONENT_NAME: cls.__name__,
            KEY_PROMPT_NAME: prompt_name,
            KEY_USER_PROMPT: user_prompt,
            KEY_LLM_RESPONSE_METADATA: llm_response.to_dict() if llm_response else None,
        }

        parse_data[PROMPTS].append(prompt_data)  # type: ignore[literal-required]

    def train(  # type: ignore[override]
        self,
        training_trackers: List[TrackerWithCachedStates],
        domain: Domain,
        responses: Responses,
        forms: Forms,
        training_data: TrainingData,
        **kwargs: Any,
    ) -> Resource:
        """Trains a policy.

        Args:
            training_trackers: The story and rules trackers from the training data.
            domain: The model's domain.
            responses: The model's responses.
            forms: The model's forms.
            training_data: The model's training data.
            **kwargs: Depending on the specified `needs` section and the resulting
                graph structure the policy can use different input to train itself.

        Returns:
            A policy must return its resource locator so that potential children nodes
            can load the policy from the resource.
        """
        # Perform health checks for both LLM and embeddings client configs
        # Skip LLM health checks if use_generative_llm is False
        self._perform_health_checks(
            self.llm_config,
            self.embeddings_config,
            "enterprise_search_policy.train",
            use_generative_llm=self.use_llm,
        )

        # telemetry call to track training start
        track_enterprise_search_policy_train_started()

        # validate embedding configuration
        try:
            embeddings = self._create_plain_embedder(self.embeddings_config)
        except (ValidationError, Exception) as e:
            raise EnterpriseSearchPolicyError(
                code="core.policies.enterprise_search_policy.train"
                ".embedder_instantiation_failed",
                event_info=f"Unable to create embedder. Please make sure you "
                f"specified the required environment variables. Error: {e}",
            )

        if self.vector_store_type == DEFAULT_VECTOR_STORE_TYPE:
            structlogger.info("enterprise_search_policy.train.faiss")
            docs_folder = self.vector_store_config.get(SOURCE_PROPERTY)
            self._validate_documents_folder(docs_folder)
            with self._model_storage.write_to(self._resource) as path:
                self.vector_store = FAISS_Store(
                    docs_folder=docs_folder,
                    embeddings=embeddings,
                    index_path=path,
                    create_index=True,
                    parse_as_faq_pairs=not self.use_llm,
                )
        else:
            structlogger.info(
                "enterprise_search_policy.train.custom",
                store_type=self.vector_store_type,
            )

        # telemetry call to track training completion
        track_enterprise_search_policy_train_completed(
            vector_store_type=self.vector_store_type,
            embeddings_type=self.embeddings_config.get(PROVIDER_CONFIG_KEY),
            embeddings_model=self.embeddings_config.get(MODEL_CONFIG_KEY)
            or self.embeddings_config.get(MODEL_NAME_CONFIG_KEY),
            embeddings_model_group_id=self.embeddings_config.get(
                MODEL_GROUP_ID_CONFIG_KEY
            ),
            llm_type=self.llm_config.get(PROVIDER_CONFIG_KEY),
            llm_model=self.llm_config.get(MODEL_CONFIG_KEY)
            or self.llm_config.get(MODEL_NAME_CONFIG_KEY),
            llm_model_group_id=self.llm_config.get(MODEL_GROUP_ID_CONFIG_KEY),
            citation_enabled=self.citation_enabled,
            relevancy_check_enabled=self.relevancy_check_enabled,
            include_date_time=self.include_date_time,
            timezone=self.timezone,
        )
        self.persist()
        return self._resource

    def persist(self) -> None:
        """Persists the policy to storage."""
        with self._model_storage.write_to(self._resource) as path:
            rasa.shared.utils.io.write_text_file(
                self.prompt_template, path / ENTERPRISE_SEARCH_PROMPT_FILE_NAME
            )
            config = self.config.copy()
            config[LLM_CONFIG_KEY] = self.llm_config
            config[EMBEDDINGS_CONFIG_KEY] = self.embeddings_config
            rasa.shared.utils.io.dump_obj_as_json_to_file(
                path / ENTERPRISE_SEARCH_CONFIG_FILE_NAME, config
            )

    def _prepare_slots_for_template(
        self, tracker: DialogueStateTracker
    ) -> List[Dict[str, str]]:
        """Prepares the slots for the template.

        Args:
            tracker: The tracker containing the conversation history up to now.

        Returns:
            The non-empty slots.
        """
        template_slots = []
        for name, slot in tracker.slots.items():
            if name not in DEFAULT_SLOT_NAMES and slot.value is not None:
                template_slots.append(
                    {
                        "name": name,
                        "value": str(slot.value),
                        "type": slot.type_name,
                    }
                )
        return template_slots

    def _connect_vector_store_or_raise(
        self, endpoints: Optional[AvailableEndpoints]
    ) -> None:
        """Connects to the vector store or raises an exception.

        Raise exceptions for the following cases:
        - The configuration is not specified
        - Unable to connect to the vector store

        Args:
            endpoints: Endpoints configuration.
        """
        config = endpoints.vector_store if endpoints else None
        if config is None and self.vector_store_type != DEFAULT_VECTOR_STORE_TYPE:
            structlogger.error(
                "enterprise_search_policy._connect_vector_store_or_raise.no_config"
            )
            raise VectorStoreConfigurationError(
                """No vector store specified. Please specify a vector
                store in the endpoints configuration"""
            )
        try:
            self.vector_store.connect(config)  # type: ignore
        except Exception as e:
            structlogger.error(
                "enterprise_search_policy._connect_vector_store_or_raise.connect_error",
                error=e,
                config=config,
            )
            raise VectorStoreConnectionError(
                f"Unable to connect to the vector store. Error: {e}"
            )

    def _prepare_search_query(self, tracker: DialogueStateTracker, history: int) -> str:
        """Prepares the search query.
        The search query is the last N messages in the conversation history.

        Args:
            tracker: The tracker containing the conversation history up to now.
            history: The number of messages to include in the search query.

        Returns:
            The search query.
        """
        transcript = []
        for event in tracker.applied_events():
            if isinstance(event, UserUttered) or isinstance(event, BotUttered):
                transcript.append(sanitize_message_for_prompt(event.text))

        search_query = " ".join(transcript[-history:][::-1])
        structlogger.debug("search_query", search_query=search_query)
        return search_query

    def get_llm_tracing_metadata(self, tracker: DialogueStateTracker) -> Dict[str, Any]:
        return {
            LANGFUSE_METADATA_SESSION_ID: tracker.sender_id,
            LANGFUSE_METADATA_TAGS: [self.__class__.__name__],
            LANGFUSE_METADATA_CUSTOM_METADATA: {
                LANGFUSE_METADATA_AGENT_ID: tracker.assistant_id,
                LANGFUSE_METADATA_MODEL_ID: tracker.model_id,
                LANGFUSE_METADATA_COMPONENT_NAME: self.__class__.__name__,
            },
        }

    async def predict_action_probabilities(  # type: ignore[override]
        self,
        tracker: DialogueStateTracker,
        domain: Domain,
        endpoints: Optional[AvailableEndpoints] = None,
        rule_only_data: Optional[Dict[Text, Any]] = None,
        **kwargs: Any,
    ) -> PolicyPrediction:
        """Predicts the next action the bot should take after seeing the tracker.

        Args:
            tracker: The tracker containing the conversation history up to now.
            domain: The model's domain.
            endpoints: The model's endpoints.
            rule_only_data: Slots and loops which are specific to rules and hence
                should be ignored by this policy.
            **kwargs: Depending on the specified `needs` section and the resulting
                graph structure the policy can use different input to make predictions.
                Includes `output_channel` for streaming responses.

        Returns:
             The prediction.
        """
        logger_key = "enterprise_search_policy.predict_action_probabilities"

        # Get output_channel from kwargs to support streaming responses.
        # This is not in the signature to avoid breaking changes.
        output_channel = kwargs.get("output_channel")

        if not self.supports_current_stack_frame(
            tracker, False, False
        ) or self.should_abstain_in_coexistence(tracker, True):
            return self._prediction(self._default_predictions(domain))

        if not self.vector_store:
            structlogger.error(f"{logger_key}.no_vector_store")
            return self._create_prediction_internal_error(domain, tracker)

        try:
            self._connect_vector_store_or_raise(endpoints)
        except (VectorStoreConfigurationError, VectorStoreConnectionError) as e:
            structlogger.error(f"{logger_key}.connection_error", error=e)
            return self._create_prediction_internal_error(domain, tracker)

        search_query = self._prepare_search_query(
            tracker, int(self.max_messages_in_query)
        )
        tracker_state = tracker.current_state(EventVerbosity.AFTER_RESTART)

        try:
            with embedding_metadata_context(self.get_llm_tracing_metadata(tracker)):
                documents = await self.vector_store.search(
                    query=search_query,
                    tracker_state=tracker_state,
                    threshold=self.vector_search_threshold,
                )
        except InformationRetrievalException as e:
            structlogger.error(f"{logger_key}.search_error", error=e)
            return self._create_prediction_internal_error(domain, tracker)

        if not documents.results:
            structlogger.info(f"{logger_key}.no_documents")
            return self._create_prediction_cannot_handle(domain, tracker)

        # Prepare the prompt for LLM
        prompt = self._render_prompt(tracker, documents.results)

        self._telemetry_policy_prediction()

        if self.use_llm:
            # Use unified method that handles both streaming and non-streaming
            return await self._return_llm_response(
                domain, tracker, search_query, prompt, documents, output_channel
            )
        else:
            # No LLM, use static response with action_send_text from search results
            return self._return_text_response(domain, tracker, search_query, documents)

    async def _return_llm_response(
        self,
        domain: Domain,
        tracker: DialogueStateTracker,
        search_query: Text,
        prompt: Text,
        documents: SearchResultList,
        output_channel: Optional["OutputChannel"],
    ) -> PolicyPrediction:
        """Returns an LLM response, with streaming support when available.

        This method handles both streaming and non-streaming cases. If output_channel
        is provided and no post-processing (relevancy check or citation) is needed,
        it will stream the response. Otherwise, it falls back to non-streaming.

        Args:
            domain: The model's domain.
            tracker: The tracker containing the conversation history up to now.
            search_query: The search query used.
            prompt: The prompt to send to the LLM.
            documents: The documents retrieved from search.
            output_channel: Optional output channel for streaming the response.

        Returns:
            The policy prediction.
        """
        logger_key = "enterprise_search_policy._return_llm_response"

        # Determine if we can use streaming (no post-processing needed)
        can_stream = (
            output_channel is not None
            and not self.relevancy_check_enabled
            and not self.citation_enabled
        )

        llm_input = LLMInput(
            prompt=prompt, metadata=self.get_llm_tracing_metadata(tracker)
        )
        if can_stream:
            llm_response = await self._invoke_llm_with_streaming(
                llm_input, output_channel, tracker.sender_id
            )
        else:
            llm_response = await self._invoke_llm(llm_input)

        self._add_prompt_and_llm_response_to_latest_message(
            tracker=tracker,
            prompt_name="enterprise_search_prompt",
            user_prompt=prompt,
            llm_response=llm_response,
        )

        if llm_response is None or not llm_response.choices:
            structlogger.debug(f"{logger_key}.no_llm_response")
            return self._create_prediction_internal_error(domain, tracker)

        llm_answer = llm_response.choices[0]

        # Apply post-processing if needed (only for non-streaming cases)
        if self.relevancy_check_enabled:
            relevancy_response = self._parse_llm_relevancy_check_response(llm_answer)
            if not relevancy_response.relevant:
                structlogger.debug(f"{logger_key}.answer_not_relevant")
                return self._create_prediction_cannot_handle(
                    domain,
                    tracker,
                    RASA_PATTERN_CANNOT_HANDLE_NO_RELEVANT_ANSWER,
                )

        if self.citation_enabled:
            llm_answer = self.post_process_citations(llm_answer)

        structlogger.debug(
            f"{logger_key}.llm_answer", prompt=prompt, llm_answer=llm_answer
        )

        action_metadata = self._get_action_metadata(search_query, documents, llm_answer)

        return self._create_prediction(
            domain=domain, tracker=tracker, action_metadata=action_metadata
        )

    def _return_text_response(
        self,
        domain: Domain,
        tracker: DialogueStateTracker,
        search_query: Text,
        documents: SearchResultList,
    ) -> PolicyPrediction:
        """Returns a static response from the search results.

        Args:
            documents: The documents retrieved from search
        Returns:
            The policy prediction with the static response.
        """
        logger_key = "enterprise_search_policy._return_static_response"
        response = documents.results[0].metadata.get("answer", None)
        if not response:
            structlogger.error(
                f"{logger_key}.answer_key_missing_in_metadata",
                search_results=documents.results,
            )
            return self._create_prediction_internal_error(domain, tracker)

        structlogger.debug(
            f"{logger_key}.no_llm",
            search_results=documents,
        )

        action_metadata = self._get_action_metadata(search_query, documents, response)

        return self._create_prediction(
            domain=domain,
            tracker=tracker,
            action_metadata=action_metadata,
        )

    def _telemetry_policy_prediction(self) -> None:
        """Sends telemetry data for policy prediction."""
        track_enterprise_search_policy_predict(
            vector_store_type=self.vector_store_type,
            embeddings_type=self.embeddings_config.get(PROVIDER_CONFIG_KEY),
            embeddings_model=self.embeddings_config.get(MODEL_CONFIG_KEY)
            or self.embeddings_config.get(MODEL_NAME_CONFIG_KEY),
            embeddings_model_group_id=self.embeddings_config.get(
                MODEL_GROUP_ID_CONFIG_KEY
            ),
            llm_type=self.llm_config.get(PROVIDER_CONFIG_KEY),
            llm_model=self.llm_config.get(MODEL_CONFIG_KEY)
            or self.llm_config.get(MODEL_NAME_CONFIG_KEY),
            llm_model_group_id=self.llm_config.get(MODEL_GROUP_ID_CONFIG_KEY),
            citation_enabled=self.citation_enabled,
            relevancy_check_enabled=self.relevancy_check_enabled,
            include_date_time=self.include_date_time,
            timezone=self.timezone,
        )

    def _get_action_metadata(
        self,
        search_query: Text,
        documents: SearchResultList,
        response: Optional[Text],
    ) -> Dict[Text, Any]:
        metadata = {
            ACTION_METADATA_MESSAGE_KEY: {
                SEARCH_RESULTS_METADATA_KEY: [
                    result.text for result in documents.results
                ],
                UTTER_SOURCE_METADATA_KEY: self.__class__.__name__,
                SEARCH_QUERY_METADATA_KEY: search_query,
            }
        }

        if response:
            metadata[ACTION_METADATA_MESSAGE_KEY][ACTION_METADATA_TEXT_KEY] = response

        return metadata

    def _render_prompt(
        self, tracker: DialogueStateTracker, documents: List[SearchResult]
    ) -> Text:
        """Renders the prompt from the template.

        Args:
            tracker: The tracker containing the conversation history up to now.
            documents: The documents retrieved from search

        Returns:
            The rendered prompt.
        """
        inputs = {
            "current_conversation": tracker_as_readable_transcript(
                tracker, max_turns=self.max_history
            ),
            "docs": documents,
            "slots": self._prepare_slots_for_template(tracker),
            "check_relevancy": self.relevancy_check_enabled,
            "citation_enabled": self.citation_enabled,
        }
        # Add current datetime if enabled
        if self.include_date_time:
            mocked_datetime_value = tracker.get_slot(MOCKED_DATETIME_SLOT)
            inputs["current_datetime"] = resolve_datetime(
                mocked_datetime_value, timezone=self.timezone
            )

        prompt = Template(self.prompt_template).render(**inputs)
        log_llm(
            logger=structlogger,
            log_module="EnterpriseSearchPolicy",
            log_event="enterprise_search_policy._render_prompt.prompt_rendered",
            prompt=prompt,
        )
        return prompt

    @measure_llm_latency
    async def _invoke_llm(self, llm_input: LLMInput) -> Optional[LLMResponse]:
        """Fetches an LLM completion for the provided prompt.

        Args:
            llm_input: The LLM input containing the prompt and metadata.

        Returns:
            An LLMResponse object, or None if the call fails.
        """
        llm = llm_factory(self.llm_config, DEFAULT_LLM_CONFIG)
        try:
            response = await llm.acompletion(
                llm_input.prompt, metadata=llm_input.metadata
            )
            return LLMResponse.ensure_llm_response(response)
        except Exception as e:
            # unfortunately, langchain does not wrap LLM exceptions which means
            # we have to catch all exceptions here
            structlogger.error(
                "enterprise_search_policy._generate_llm_answer.llm_error",
                error=e,
            )
            return None

    async def _invoke_llm_with_streaming(
        self, llm_input: LLMInput, output_channel: "OutputChannel", recipient_id: str
    ) -> Optional[LLMResponse]:
        """Fetches an LLM completion for the provided prompt with streaming support.

        Args:
            llm_input: The LLM input containing the prompt and metadata.
            output_channel: Output channel for streaming the response.
            recipient_id: Recipient ID for the output channel.

        Returns:
            An LLMResponse object, or None if the call fails.
        """
        llm = llm_factory(self.llm_config, DEFAULT_LLM_CONFIG)
        try:
            response = await acompletion_with_streaming(
                llm_client=llm,
                messages=llm_input.prompt,
                output_channel=output_channel,
                recipient_id=recipient_id,
                metadata=llm_input.metadata,
            )
            return LLMResponse.ensure_llm_response(response)
        except Exception as e:
            structlogger.error(
                "enterprise_search_policy._invoke_llm_with_streaming.error",
                error=e,
            )
            return None

    def _parse_llm_relevancy_check_response(
        self, llm_answer: str
    ) -> _RelevancyCheckResponse:
        """Checks if the LLM response is relevant by parsing it."""
        answer_relevant = not _ENTERPRISE_SEARCH_ANSWER_NOT_RELEVANT_PATTERN.search(
            llm_answer
        )

        return _RelevancyCheckResponse(
            answer=llm_answer if answer_relevant else None,
            relevant=answer_relevant,
        )

    def _create_prediction(
        self,
        domain: Domain,
        tracker: DialogueStateTracker,
        action_metadata: Dict[Text, Any],
    ) -> PolicyPrediction:
        """Create a policy prediction result with the given action name.

        Args:
            domain: The model's domain.
            tracker: The tracker containing the conversation history up to now.
            action_metadata: The metadata for the predicted action.

        Returns:
            The prediction.
        """
        result = self._prediction_result(ACTION_SEND_TEXT_NAME, domain)
        stack = tracker.stack
        if not stack.is_empty():
            stack.pop()
            events: List[Event] = tracker.create_stack_updated_events(stack)
        else:
            events = []

        return self._prediction(result, action_metadata=action_metadata, events=events)

    def _create_prediction_internal_error(
        self, domain: Domain, tracker: DialogueStateTracker
    ) -> PolicyPrediction:
        return self._create_prediction_for_pattern(
            domain, tracker, InternalErrorPatternFlowStackFrame()
        )

    def _create_prediction_cannot_handle(
        self,
        domain: Domain,
        tracker: DialogueStateTracker,
        reason: Optional[str] = None,
    ) -> PolicyPrediction:
        cannot_handle_stack_frame = (
            CannotHandlePatternFlowStackFrame(reason=reason)
            if reason is not None
            else CannotHandlePatternFlowStackFrame()
        )
        return self._create_prediction_for_pattern(
            domain, tracker, cannot_handle_stack_frame
        )

    def _create_prediction_for_pattern(
        self,
        domain: Domain,
        tracker: DialogueStateTracker,
        pattern_stack_frame: PatternFlowStackFrame,
    ) -> PolicyPrediction:
        """Create a policy prediction result for error.

        We should cancel the current flow (hence ACTION_CANCEL_FLOW) and push a
        pattern stack frame (Internal Error Pattern by default) to start the pattern.

        Args:
            domain: The model's domain.
            tracker: The tracker containing the conversation history up to now.
            pattern_stack_frame: The pattern stack frame to push.

        Returns:
            The prediction.
        """
        # TODO: replace ACTION_CANCEL_FLOW (ATO-2097)
        result = self._prediction_result(ACTION_CANCEL_FLOW, domain)
        stack = tracker.stack
        if not stack.is_empty():
            stack.pop()
            stack.push(pattern_stack_frame)
        events: List[Event] = tracker.create_stack_updated_events(stack)
        return self._prediction(result, action_metadata=None, events=events)

    def _prediction_result(
        self, action_name: Optional[Text], domain: Domain, score: Optional[float] = 1.0
    ) -> List[float]:
        """Creates a prediction result.

        Args:
            action_name: The name of the predicted action.
            domain: The model's domain.
            score: The score of the predicted action.

        Returns:
        The prediction result where the score is used for one hot encoding.
        """
        result = self._default_predictions(domain)
        if action_name:
            result[domain.index_for_action(action_name)] = score  # type: ignore[assignment]
        return result

    @classmethod
    def _validate_documents_folder(cls, docs_folder: str) -> None:
        if not os.path.exists(docs_folder) or not os.path.isdir(docs_folder):
            error_message = (
                f"Document source directory does not exist or is not a "
                f"directory: '{os.path.abspath(docs_folder)}'. "
                "Please specify a valid path to the documents source directory in the "
                "vector_store configuration."
            )
            error_code = (
                "core.policies.enterprise_search_policy.train"
                ".faiss.invalid_source_directory"
            )
            raise EnterpriseSearchPolicyError(code=error_code, event_info=error_message)

        docs = glob.glob(os.path.join(docs_folder, "**", "*.txt"), recursive=True)
        if not docs or len(docs) < 1:
            error_message = (
                f"Document source directory is empty: '{docs_folder}'. "
                "Please add documents to this directory or specify a different one."
            )
            error_code = (
                "core.policies.enterprise_search_policy.train"
                ".faiss.source_directory_empty"
            )
            raise EnterpriseSearchPolicyError(code=error_code, event_info=error_message)

    @classmethod
    def load(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
        **kwargs: Any,
    ) -> "EnterpriseSearchPolicy":
        """Loads a trained policy (see parent class for full docstring)."""
        parsed_config = EnterpriseSearchPolicyConfig.from_dict(config)

        # Perform health checks for both LLM and embeddings client configs
        # Skip LLM health checks if use_generative_llm is False
        cls._perform_health_checks(
            parsed_config.llm_config,
            parsed_config.embeddings_config,
            "enterprise_search_policy.load",
            use_generative_llm=parsed_config.use_generative_llm,
        )

        prompt_template = cls._load_prompt_template(model_storage, resource)
        embeddings = cls._create_plain_embedder(parsed_config.embeddings_config)
        vector_store = cls._load_vector_store(
            embeddings,
            parsed_config.vector_store_type,
            parsed_config.use_generative_llm,
            model_storage,
            resource,
        )

        structlogger.info("enterprise_search_policy.load", config=config)

        return cls(
            config,
            model_storage,
            resource,
            execution_context,
            vector_store=vector_store,
            prompt_template=prompt_template,
        )

    @classmethod
    def _load_prompt_template(
        cls, model_storage: ModelStorage, resource: Resource
    ) -> Optional[str]:
        try:
            with model_storage.read_from(resource) as path:
                return rasa.shared.utils.io.read_file(
                    path / ENTERPRISE_SEARCH_PROMPT_FILE_NAME
                )
        except (FileNotFoundError, FileIOException) as e:
            structlogger.warning(
                "enterprise_search_policy.load.failed", error=e, resource=resource.name
            )
        return None

    @classmethod
    def _load_vector_store(
        cls,
        embeddings: "Embeddings",
        store_type: str,
        use_generative_llm: bool,
        model_storage: ModelStorage,
        resource: Resource,
    ) -> InformationRetrieval:
        if store_type == DEFAULT_VECTOR_STORE_TYPE:
            # if a vector store is not specified,
            # default to using FAISS with the index stored in the model
            # TODO figure out a way to get path without context manager
            with model_storage.read_from(resource) as path:
                return FAISS_Store(
                    embeddings=embeddings,
                    index_path=path,
                    docs_folder=None,
                    create_index=False,
                    parse_as_faq_pairs=not use_generative_llm,
                )
        else:
            return create_from_endpoint_config(
                config_type=store_type,
                embeddings=embeddings,
            )

    @classmethod
    def _get_local_knowledge_data(
        cls, store_type: str, source: Optional[str] = None
    ) -> Optional[List[str]]:
        """This is required only for local knowledge base types.

        e.g. FAISS, to ensure that the graph component is retrained when the knowledge
        base is updated.
        """
        if (
            store_type != DEFAULT_VECTOR_STORE_TYPE
            or not source
            or not os.path.exists(source)
            or not os.path.isdir(source)
        ):
            return None

        docs = FAISS_Store.load_documents(source)

        if len(docs) == 0:
            return None

        docs_as_strings = [
            json.dumps(doc.dict(), ensure_ascii=False, sort_keys=True) for doc in docs
        ]
        return sorted(docs_as_strings)

    @classmethod
    def fingerprint_addon(cls, config: Dict[str, Any]) -> Optional[str]:
        """Add a fingerprint of enterprise search policy for the graph."""
        parsed_config = EnterpriseSearchPolicyConfig.from_dict(config)

        # Resolve the prompt template
        default_prompt_template = cls._select_default_prompt_template_based_on_features(
            parsed_config.check_relevancy, parsed_config.enable_citation
        )
        prompt_template = get_prompt_template(
            jinja_file_path=parsed_config.prompt_template,
            default_prompt_template=default_prompt_template,
            log_source_component=EnterpriseSearchPolicy.__name__,
            log_source_method=LOG_COMPONENT_SOURCE_METHOD_FINGERPRINT_ADDON,
        )

        # Fetch the local knowledge data in case FAISS is used
        local_knowledge_data = cls._get_local_knowledge_data(
            parsed_config.vector_store_type, parsed_config.vector_store_source
        )

        # Get the resolved LLM and embeddings configurations
        llm_config = parsed_config.llm_config
        embedding_config = parsed_config.embeddings_config

        return deep_container_fingerprint(
            [prompt_template, local_knowledge_data, llm_config, embedding_config]
        )

    @staticmethod
    def post_process_citations(llm_answer: str) -> str:
        """Post-processes the LLM answer to correctly number and sort citations and
        sources.

        - Handles both single `[1]` and grouped `[1, 3]` citations.
        - Rewrites the numbers in square brackets in the answer text to start from 1
        and be sorted within each group.
        - Reorders the sources according to the order of their first appearance
        in the text.
        - Removes citations from the text that point to sources missing from
        the source list.
        - Keeps sources that are not cited in the text, placing them at the end
        of the list.

        Args:
            llm_answer: The LLM answer.

        Returns:
            The post-processed LLM answer.
        """
        structlogger.debug(
            "enterprise_search_policy.post_process_citations", llm_answer=llm_answer
        )

        # Split llm_answer into answer and citations
        try:
            answer_part, sources_part = llm_answer.rsplit("Sources:", 1)
        except ValueError:
            # if there is no "Sources:" separator, return the original llm_answer
            return llm_answer.strip()

        # Parse the sources block to extract valid sources and other lines
        valid_sources, other_source_lines = EnterpriseSearchPolicy._parse_sources_block(
            sources_part
        )

        # Find all unique, valid citations in the answer text in their order
        # of appearance
        cited_order = EnterpriseSearchPolicy._get_cited_order(
            answer_part, valid_sources
        )

        # Create a mapping from the old source numbers to the new, sequential numbers.
        # For example, if the citation order in the text was [3, 1, 2], this map
        # becomes {3: 1, 1: 2, 2: 3}. This allows for a quick lookup when rewriting
        # the citations
        renumbering_map = {
            old_num: new_num + 1 for new_num, old_num in enumerate(cited_order)
        }

        # Rewrite the citations in the answer text based on the renumbering map
        processed_answer = EnterpriseSearchPolicy._rewrite_answer_citations(
            answer_part, renumbering_map
        )

        # Build the new list of sources
        new_sources_list = EnterpriseSearchPolicy._build_final_sources_list(
            cited_order,
            renumbering_map,
            valid_sources,
            other_source_lines,
        )

        if len(new_sources_list) > 0:
            processed_answer += "\nSources:\n" + "\n".join(new_sources_list)

        return processed_answer

    @staticmethod
    def _parse_sources_block(sources_part: str) -> Tuple[Dict[int, str], List[str]]:
        """Parses the sources block from the LLM response.
        Returns a tuple containing:
        - A dictionary of valid sources matching the "[1] ..." format,
        where the key is the source number
        - A list of other source lines that do not match the specified format
        """
        valid_sources: Dict[int, str] = {}
        other_source_lines: List[str] = []
        source_line_pattern = re.compile(r"^\s*\[(\d+)\](.*)")

        source_lines = sources_part.strip().split("\n")

        for line in source_lines:
            line = line.strip()
            if not line:
                continue

            match = source_line_pattern.match(line)
            if match:
                num = int(match.group(1))
                valid_sources[num] = line
            else:
                other_source_lines.append(line)

        return valid_sources, other_source_lines

    @staticmethod
    def _get_cited_order(
        answer_part: str, available_sources: Dict[int, str]
    ) -> List[int]:
        """Find all unique, valid citations in the answer text in their order
        # of appearance
        """
        cited_order: List[int] = []
        seen_indices = set()

        for match in _ENTERPRISE_SEARCH_CITATION_PATTERN.finditer(answer_part):
            content = match.group(1)
            indices_str = [s.strip() for s in content.split(",")]
            for index_str in indices_str:
                if index_str.isdigit():
                    index = int(index_str)
                    if index in available_sources and index not in seen_indices:
                        cited_order.append(index)
                        seen_indices.add(index)

        return cited_order

    @staticmethod
    def _rewrite_answer_citations(
        answer_part: str, renumber_map: Dict[int, int]
    ) -> str:
        """Rewrites the citations in the answer text based on the renumbering map."""

        def replacer(match: re.Match) -> str:
            content = match.group(1)
            old_indices_str = [s.strip() for s in content.split(",")]
            new_indices = [
                renumber_map[int(s)]
                for s in old_indices_str
                if s.isdigit() and int(s) in renumber_map
            ]
            if not new_indices:
                return ""

            return f"[{', '.join(map(str, sorted(list(set(new_indices)))))}]"

        processed_answer = _ENTERPRISE_SEARCH_CITATION_PATTERN.sub(
            replacer, answer_part
        )

        # Clean up formatting after replacements
        processed_answer = re.sub(r"\s+([,.?])", r"\1", processed_answer)
        processed_answer = processed_answer.replace("[]", " ")
        processed_answer = re.sub(r"\s+", " ", processed_answer)
        processed_answer = processed_answer.strip()

        return processed_answer

    @staticmethod
    def _build_final_sources_list(
        cited_order: List[int],
        renumbering_map: Dict[int, int],
        valid_sources: Dict[int, str],
        other_source_lines: List[str],
    ) -> List[str]:
        """Builds the final list of sources based on the cited order and
        renumbering map.
        """
        new_sources_list: List[str] = []

        # First, add the sorted, used sources
        for old_num in cited_order:
            new_num = renumbering_map[old_num]
            source_line = valid_sources[old_num]
            new_sources_list.append(
                source_line.replace(f"[{old_num}]", f"[{new_num}]", 1)
            )

        # Then, add the unused but validly numbered sources
        used_source_nums = set(cited_order)
        # Sort by number to ensure a consistent order for uncited sources
        for num, line in sorted(valid_sources.items()):
            if num not in used_source_nums:
                new_sources_list.append(line)

        # Finally, add any other source lines
        new_sources_list.extend(other_source_lines)

        return new_sources_list

    @classmethod
    def _perform_health_checks(
        cls,
        llm_config: Dict[Text, Any],
        embeddings_config: Dict[Text, Any],
        log_source_method: str,
        use_generative_llm: bool = True,
    ) -> None:
        """Perform the health checks using resolved LLM and embeddings configurations.
        Resolved means the configuration is either:
        - A reference to a model group that has already been expanded into
          its corresponding configuration using the information from
          `endpoints.yml`, or
        - A full configuration for the embedder defined directly (i.e. not
          relying on model groups or indirections).

        Args:
            llm_config: A resolved LLM configuration.
            embeddings_config: A resolved embeddings configuration.
            log_source_method: The method health checks has been called from.
            use_generative_llm: Whether to perform LLM health checks. If False,
                only embeddings health checks are performed.

        """
        # Only perform LLM health checks if use_generative_llm is True
        if use_generative_llm:
            cls.perform_llm_health_check(
                llm_config,
                DEFAULT_LLM_CONFIG,
                log_source_method,
                EnterpriseSearchPolicy.__name__,
            )
        cls.perform_embeddings_health_check(
            embeddings_config,
            DEFAULT_EMBEDDINGS_CONFIG,
            log_source_method,
            EnterpriseSearchPolicy.__name__,
        )

    @classmethod
    def get_system_default_prompt_based_on_config(cls, config: Dict[str, Any]) -> str:
        """Resolves the default prompt template for Enterprise Search Policy based on
        the component's configuration.

        - The old prompt is selected when both citation and relevancy check are either
          disabled or not set in the configuration.
        - The citation prompt is used when citation is enabled and relevancy check is
          either disabled or not set in the configuration.
        - The relevancy check prompt is only used when relevancy check is enabled.

        Args:
            config: The component's configuration.

        Returns:
            The resolved jinja prompt template as a string.
        """
        # Get the feature flags
        parsed_config = EnterpriseSearchPolicyConfig.from_dict(config)
        # Based on the enabled features (citation, relevancy check) fetch the
        # appropriate default prompt
        default_prompt = cls._select_default_prompt_template_based_on_features(
            parsed_config.check_relevancy, parsed_config.enable_citation
        )

        return default_prompt

    @classmethod
    def _select_default_prompt_template_based_on_features(
        cls,
        relevancy_check_enabled: bool,
        citation_enabled: bool,
    ) -> str:
        """Returns the appropriate default prompt template based on the feature flags.

        The selection follows this priority:
        1. If relevancy check is enabled, return the prompt that includes both relevancy
           and citation blocks.
        2. If only citation is enabled, return the prompt with citation blocks.
        3. Otherwise, fall back to the legacy default prompt template.

        Args:
            relevancy_check_enabled: Whether the LLM-generated answer should undergo
                relevancy evaluation.
            citation_enabled: Whether citations should be included in the generated
                answer.

        Returns:
            The default prompt template corresponding to the enabled features.
        """
        if relevancy_check_enabled:
            # ES prompt that has relevancy check and citations blocks
            return DEFAULT_ENTERPRISE_SEARCH_PROMPT_WITH_RELEVANCY_CHECK_AND_CITATION_TEMPLATE  # noqa: E501
        elif citation_enabled:
            # ES prompt with citation's block - backward compatibility
            return DEFAULT_ENTERPRISE_SEARCH_PROMPT_WITH_CITATION_TEMPLATE
        else:
            # Legacy ES prompt - backward compatibility
            return DEFAULT_ENTERPRISE_SEARCH_PROMPT_TEMPLATE



from typing import Dict, List, Tuple

from rasa.dialogue_understanding_test.du_test_case import (
    DialogueUnderstandingTestCase,
    DialogueUnderstandingTestStep,
)
from rasa.e2e_test.e2e_test_case import Metadata
from rasa.e2e_test.e2e_test_runner import E2ETestRunner


def filter_metadata(
    test_case: DialogueUnderstandingTestCase,
    user_step: DialogueUnderstandingTestStep,
    metadata: List[Metadata],
    sender_id: str,
) -> Dict[str, str]:
    """Filter metadata for a test case and a step."""
    # test case metadata
    test_case_metadata = E2ETestRunner.filter_metadata_for_input(
        test_case.metadata_name, metadata
    )
    test_case_metadata_dict = test_case_metadata.metadata if test_case_metadata else {}

    # step metadata
    step_metadata = E2ETestRunner.filter_metadata_for_input(
        user_step.metadata_name, metadata
    )
    step_metadata_dict = step_metadata.metadata if step_metadata else {}

    # merge metadata
    return E2ETestRunner.merge_metadata(
        sender_id, user_step.text, test_case_metadata_dict, step_metadata_dict
    )


def get_command_comparison(step: DialogueUnderstandingTestStep) -> List[str]:
    expected_commands = (
        [command.to_dsl() for command in step.commands] if step.commands else []
    )
    predicted_commands = [command.to_dsl() for command in step.get_predicted_commands()]

    expected_commands.insert(0, "---EXPECTED---")
    predicted_commands.insert(0, "---PREDICTED---")

    max_line_length = max(len(line) for line in expected_commands)
    expected_commands, predicted_commands = make_lists_equal(
        expected_commands, predicted_commands
    )

    command_comparison = []
    for i, (expected_line, actual_line) in enumerate(
        zip(expected_commands, predicted_commands)
    ):
        expected_line += " " * (max_line_length - len(expected_line))
        if i == 0:
            # make the first line red:
            expected_line = "[red3]" + expected_line + "[/red3]"
            actual_line = "[red3]" + actual_line + "[/red3]"
        command_comparison.append(f"{expected_line} | {actual_line}")

    return command_comparison


def make_lists_equal(list1: List[str], list2: List[str]) -> Tuple[List[str], List[str]]:
    if len(list1) < len(list2):
        list1.extend([""] * (len(list2) - len(list1)))
        return list1, list2

    list2.extend([""] * (len(list1) - len(list2)))
    return list1, list2
